import pathlib
import tempfile
import omidb
import pytest
from collections import namedtuple


Dirs = namedtuple("Dirs", "root omidb data images")


@pytest.fixture
def dirs():

    with tempfile.TemporaryDirectory() as root_dir:
        root = pathlib.Path(root_dir)
        omidb_dir = root / "omidb"
        data_dir = omidb_dir / "data"
        images_dir = omidb_dir / "images"

        omidb_dir.mkdir()
        data_dir.mkdir()
        images_dir.mkdir()

        yield Dirs(root, omidb_dir, data_dir, images_dir)


def test_missing_data_dir_raise(dirs: Dirs):
    dirs.data.rmdir()
    with pytest.raises(FileNotFoundError) as e:
        assert omidb.DB(dirs.data / "nodir")


def test_find_client_dirs(dirs: Dirs):
    clients = set(["demd1", "demd2", "optm3", "nonsense"])
    for client in clients:
        (dirs.data / client).mkdir()

    parser = omidb.DB(dirs.data)
    clients.remove("nonsense")
    assert parser.clients == clients


def test_parse_study_dir(monkeypatch, mocker, dirs: Dirs):
    client = "demd1"
    studies = ["1.2.3.4567", "nonsense"]
    for study in studies:
        study_dir = dirs.data / "demd1" / study
        study_dir.mkdir(parents=True)
        print(study_dir)

    def mock(self, client, studies):
        return client, set(studies)

    mocker.patch("omidb.DB._parse_client", mock)
    parser = omidb.DB(dirs.data)
    expected = set([studies[0]])
    for parsed_client, parsed_studies in parser:
        assert client == parsed_client
        assert expected == parsed_studies



from typing import Optional, List
from dataclasses import dataclass
import datetime
from .events import Event
from .series import Series


@dataclass
class Study:
    """
    An imaging examination comprising a number of series, each containing a set
    of images.

    :param id: Study Instance UID, a unique identifier
    :param series: A list of :class:`omidb.series.Series`' containing a set of
        mammograms
    :param date: Date of the examination
    :param event_type: Enumerations defining the medical procedures associated
        with the study
    """

    id: str
    series: List[Series]
    date: Optional[datetime.date] = None
    event_type: Optional[List[Event]] = None



import enum
import datetime
import re
from typing import Optional, Union, Dict, Any, List
from .image import Image
import json
from loguru import logger

IUID_REG = re.compile(r"[\d+.]+$")


def str_to_date(date: str) -> datetime.date:
    if "-" in date:
        return datetime.datetime.strptime(date, "%Y-%m-%d").date()
    else:
        return datetime.datetime.strptime(date, "%Y%m%d").date()


def date_or_none(d: Dict[Any, Any], key: Any) -> Optional[datetime.date]:
    if d.get(key):
        return str_to_date(d[key])
    return None


def enum_lookup(
    value: str, e: enum.EnumMeta, should_raise: bool = False
) -> Optional[Any]:
    if value is None or value == "":
        return None
    for mem in e:  # type: ignore
        if value == mem.name or value == mem.value:  # type: ignore
            return mem
    logger.warning(f"`{value}` was not found in enum `{e}`")
    if should_raise:
        raise ValueError(f"`{value}` was not found in enum `{e}`")
    return None


def try_image_attribute(
    attributes: Union[Dict[str, Any], Image], tags: Union[str, List[str]]
) -> Optional[Any]:
    if isinstance(tags, str):
        tags = [tags]

    try:
        if isinstance(attributes, Image):
            attrs = attributes.attributes
        else:
            attrs = attributes
        for t in tags:
            attrs = attrs[t]["Value"][0]  # type: ignore
        return attrs
    except (KeyError, json.decoder.JSONDecodeError):
        pass
    return None


def parse_age(image: Image) -> Optional[int]:
    try:
        age_s = try_image_attribute(image, "00101010")
    except AttributeError:
        return None
    if not isinstance(age_s, str):
        return None
    match = re.match(r"\d+", age_s.lstrip("0"))
    if match is None:
        return None
    age = int(match.group())
    if age_s.endswith("M"):
        age = age // 12
    if age_s.endswith("W"):
        age = age // 52
    if age_s.endswith("D"):
        age = age // 365
    return age


# TODO: Replace these with enums
ListProperties = {
    " ": (
        "ReferralReasons",
        "IntervalCancerReviewers",
        "IntervalCancerReviewersN",
        "AdditionalBiopsyProcs",
        "AdditionalTreatmentProcs",
        "BenignLesions",
        "EpithelialProliferation",
        "Invasive",
        "NonInvasive",
        "RadiotherapySites",
        "NewTreatments",
        "Clinicians",
        "CliniciansNational",
        "MammoAttributes",
        "MammoClinicians",
        "MammoCliniciansNat",
        "UssAttributes",
        "UssClinicians",
        "UssCliniciansNat",
        "Attendees",
        "AttendeesNational",
        "OutcomeSubcategory",
        "Readers",
        "RecommendedActions",
        "FilmReaders",
        "FilmRecommendedActions",
        "Appearance",
        "Margin",
        "ViewsProcedures",
        "AdditionalDiagnosticProcs",
        "AdditionalTreatmentProcs",
        "AxillarySpecimenTypes",
        "BenignLesions",
        "DcisGrowthPatterns",
        "DiseaseExtent",
        "EpithelialProliferation",
        "InSituComponents",
        "InvasiveComponents",
        "NonsurgicalTreatments",
        "InSituComponents",
        "UncertainLesions",
    ),
    ",": (
        "SpecialAppointmentReasons",
        "OtherGeneCodes",
        "EpisodeEndPoint",
        "ReadersNational",
        "FilmReadersNational",
    ),
}


# TODO: Drop the need for `key`
def nbss_str_to_enum(
    value: Any, e: enum.EnumMeta, key: Optional[str] = None
) -> Optional[Any]:
    if value in (None, ""):
        return None

    # Check if property is 'list like' and attempt to parse as list of enums
    if key is not None:
        for sep, props in ListProperties.items():
            if key in props:
                result = []
                for v in value.split(sep):
                    t = enum_lookup(v, e)
                    if t is not None:
                        result.append(t)
                return result

    # Single value, so return enum
    return enum_lookup(value, e)



import re
import pathlib
import json
from typing import List, Dict, Optional, Iterator, Any, Sequence, Union
from loguru import logger
import pydicom
from .image import LoaderParams
from .client import Client
from .client_parser import ClientParser


class DB:
    """
    OMI-DB parser

    :param data_dir: Root directory of the OMI-DB data directory.
    :param image_dir: Root directory of the OMI-DB image directory.
    :param ignore_missing_images: If ``True``, the existence of dicom images
        belonging to a series will not be checked: parsing is based entirely on the
        JSON representations of the DICOM headers. Set to ``False`` to parse only
        those images for which you have *both* JSON and DICOM files for.
    :param clients: Only parse these clients, if they exist
    :param exclude_clients: Exclude these clients, even if they are in ``clients``
    :param distinct_event_study_links: Only match events to imaging studies
        when distinct (1 to 1 mapping)
    :param nbss_dir: An alternative data dir where nbss files can be found
    """

    def __init__(
        self,
        data_dir: Union[str, pathlib.Path],
        image_dir: Optional[Union[str, pathlib.Path]] = None,
        ignore_missing_images: bool = True,
        clients: Optional[Sequence[str]] = None,
        exclude_clients: Optional[Sequence[str]] = None,
        distinct_event_study_links: bool = True,
        nbss_dir: Optional[Union[str, pathlib.Path]] = None,
    ):
        self.alternative_nbss_dir = None if nbss_dir is None else pathlib.Path(nbss_dir)
        self.ignore_missing_images = ignore_missing_images
        self.distinct_event_study_links = distinct_event_study_links

        self._image_dir = (
            pathlib.Path() if image_dir is None else pathlib.Path(image_dir)
        )
        self._data_dir = pathlib.Path(data_dir)

        if not self._data_dir.is_dir():
            raise FileNotFoundError(f"Directory {data_dir} not found")

        if not clients:
            clients = []
            for client_path in self._data_dir.glob("*"):
                if client_path.is_dir() and client_path.name[:4] in ("demd", "optm"):
                    clients.append(client_path.name)

        self.clients = set(clients)

        if exclude_clients:
            self.clients = set(clients) - set(exclude_clients)

    def __iter__(self) -> Iterator[Client]:
        """
        Iterates over all parsable clients found in the OMI-DB directory.

        :return: client_it: A :class:`omidb.client.Client` iterator
        """
        for client in self.clients:
            _data_dir = self._data_dir / client
            studies: List[str] = []
            for study in _data_dir.glob("*/**"):
                # Extract study ID from path
                if study.is_dir():
                    match = re.match(r"\d+.", study.name)
                    if match:
                        studies.append(study.name)

            try:
                yield self._parse_client(client, studies)
            except Exception:
                logger.exception(f"Failed to parse {client}, skipping")
                continue

    def _parse_client(self, client_id: str, studies: List[str]) -> Client:
        imagedb = self._imagedb(client_id)
        nbss = self._nbss(client_id)

        client = ClientParser(
            client_id,
            nbss,
            imagedb,
            studies,
            self.distinct_event_study_links,
            self._json_loader,
            self._dcm_loader,
        )()

        return client

    def _nbss_path(self, client_id: str) -> pathlib.Path:
        """Path of the NBSS json file corresponding to the client with ID
        `client_id`
        """

        if self.alternative_nbss_dir is not None:
            data_dir = self.alternative_nbss_dir
        else:
            data_dir = self._data_dir

        p1 = data_dir / client_id / ("nbss_" + client_id + ".json")

        if not p1.exists():
            p1 = data_dir / client_id / ("NBSS_" + client_id + ".json")

        return p1

    def _nbss(self, client_id: str) -> Dict[str, Any]:
        """NBSS data corresponding to the client with ID `client_id`"""

        with open(self._nbss_path(client_id)) as f:
            nbss: Dict[str, Any] = json.load(f)
        return nbss

    def _imagedb(self, client_id: str) -> Dict[str, Any]:
        """IMAGEDB data corresponding to the client with ID `client_id`"""

        p1 = self._data_dir / client_id / ("imagedb_" + client_id + ".json")

        if not p1.exists():
            p1 = self._data_dir / client_id / ("IMAGEDB_" + client_id + ".json")

        with open(p1) as f:
            imagedb: Dict[str, Any] = json.load(f)

        return imagedb

    def _dcm_loader(self, p: LoaderParams) -> pydicom.dataset.FileDataset:
        dcm_path = self._image_dir / p.client_id / p.study_id / (p.image_id + ".dcm")
        return pydicom.dcmread(str(dcm_path))

    def _json_loader(self, p: LoaderParams) -> Dict[str, Any]:
        json_path = self._data_dir / p.client_id / p.study_id / (p.image_id + ".json")
        # Due to inconsistency in file naming
        if not json_path.exists():
            json_path = json_path.with_suffix(".dcm.json")

        result: Dict[str, Any] = {}
        with open(json_path) as f:
            result = json.load(f)
        return result



import datetime
import pathlib
import omidb
from omidb.classificationtools import _earliest_date
from typing import Any


def get_client() -> omidb.client.Client:

    mark = omidb.mark.Mark(
        "1", "2", (omidb.mark.BoundingBox(0, 0, 0, 0)
                   ), omidb.mark.Conspicuity.subtle
    )

    image = omidb.image.Image("1.2.3", pathlib.Path(), pathlib.Path(), [mark])

    # Make up some attributes for this dicom
    image._json = {
        "00080070": {"Value": ["HOLOGIC, Inc."]},
        "00081090": {"Value": ["New Model"]},
        "00080068": {"Value": ["FOR PROCESSING"]},
        "00200062": {"Value": ["L"]},
        "00185101": {"Value": ["CC"]},
        "00540220": {
            "Value": [
                {
                    "00080100": {"Value": ["R-10242"]},
                    "00080104": {"Value": ["cranio-caudal"]},
                }
            ],
        },
        "001811A0": {"Value": [50]},
        "00101010": {"Value": ["065Y"]},
        "00020010": {"Value": ["1.2.840.10008.1.2"]},
    }

    series = omidb.series.Series("1.2.3.4", [image])

    study = omidb.study.Study(
        "1.2.3.4.5",
        series=[series],
        event_type=[omidb.events.Event.screening, omidb.events.Event.clinical],
        date=datetime.date(2014, 6, 12),
    )

    events = omidb.events.Events(
        screening=[omidb.events.Screening(dates=[datetime.date(2000, 2, 1)])],
        clinical=omidb.events.BaseEvent(
            left_opinion=omidb.events.SideOpinion.ON),
    )

    episode = omidb.episode.Episode(
        "1.2.3.4.5.6",
        studies=[study],
        events=events,
        type=omidb.episode.Type.F,
        action=omidb.episode.Action.EC,
        opened_date=datetime.date(2014, 6, 12),
        closed_date=datetime.date(2014, 8, 12),
    )

    client = omidb.client.Client("abc1", [episode], "site")

    return client


def test_write_client_data() -> None:

    client = get_client()
    writer = omidb.commands.summarise.write_client_data([client])

    expected: Any = None
    # type: ignore
    for k, v, v2 in zip(writer.rows[0], writer.rows[1], writer.rows[2]):
        if k == "ClientID":
            expected = client.id
        elif k == "ClientStatus":
            expected = client.status.name
        elif k == "Site":
            expected = client.site
        elif k == "EpisodeID":
            expected = client.episodes[0].id
        elif k == "EpisodeSortDate":
            expected = _earliest_date(client.episodes[0])
        elif k == "EpisodeType":
            expected = client.episodes[0].type.name
        elif k == "EpisodeStatus":
            expected = omidb.episode.Status.N.name
        elif k == "EpisodeOutcome":
            expected = "N"
        elif k == "EpisodeOutcomeFutureEpisodeID":
            expected = None
        elif k == "EpisodeIsPostOp":
            expected = False
        elif k == "EpisodeAction":
            expected = client.episodes[0].action.name
        elif k == "EpisodeContainsMalignantOpinions":
            expected = client.episodes[0].has_malignant_opinions
        elif k == "EpisodeContainsBenignOpinions":
            expected = client.episodes[0].has_benign_opinions
        elif k == "EpisodeOpenedDate":
            expected = client.episodes[0].opened_date
        elif k == "EpisodeClosedDate":
            expected = client.episodes[0].closed_date
        elif k == "ActualEpisodeOpenedYear":
            expected = client.episodes[0].actual_opened_year
        elif k == "EpisodeHasEvents":
            expected = client.episodes[0].events is not None
        elif k == "StudyInstanceUID":
            expected = client.episodes[0].studies[0].id
        elif k == "StudyDate":
            expected = client.episodes[0].studies[0].date
        elif k == "EventType":
            # type: ignore
            expected = client.episodes[0].studies[0].event_type[0].name
            assert v == expected
            # type: ignore
            expected = client.episodes[0].studies[0].event_type[1].name
            assert v2 == expected
            continue
        elif k == "SeriesInstanceUID":
            expected = client.episodes[0].studies[0].series[0].id
        elif k == "SeriesInstanceUID":
            expected = client.episodes[0].studies[0].series[0].id
        elif k == "SOPInstanceUID":
            expected = client.episodes[0].studies[0].series[0].images[0].id
        elif k == "NumberOfMarks":
            expected = len(
                # type: ignore
                client.episodes[0].studies[0].series[0].images[0].marks
            )
        else:
            # type: ignore
            image = client.episodes[0].studies[0].series[0].images[0]
            if k == "Manufacturer":
                # type: ignore
                expected = image.attributes["00080070"]["Value"][0]
            elif k == "Model":
                # type: ignore
                expected = image.attributes["00081090"]["Value"][0]
            elif k == "BodyPartThicknessMM":
                # type: ignore
                expected = image.attributes["001811A0"]["Value"][0]
            elif k == "PatientAgeYears":
                expected = 65
            elif k == "PresentationIntentType":
                # type: ignore
                expected = image.attributes["00080068"]["Value"][0]
            elif k == "ImageLaterality":
                # type: ignore
                expected = image.attributes["00200062"]["Value"][0]
            elif k == "ViewPosition":
                # type: ignore
                expected = image.attributes["00185101"]["Value"][0]
            elif k == "ViewModCodeValue":
                expected = image.attributes["00540220"]["Value"][0]["00080100"][
                    "Value"
                ][
                    0
                ]  # type: ignore
            elif k == "ViewModCodeMeaning":
                expected = image.attributes["00540220"]["Value"][0]["00080104"][
                    "Value"
                ][
                    0
                ]  # type: ignore
            elif k == "TransferSyntaxUID":
                # type: ignore
                expected = image.attributes["00020010"]["Value"][0]
            else:
                raise ValueError(f"{k} unknown")

        assert (v == expected) & (v2 == expected)



import omidb
from .events import Event
from .episode import Episode, Type, Status, _validate_events
import copy
from typing import Optional, List, Union, Tuple
import datetime
import dataclasses
import enum
from math import ceil


def has_prior(client: omidb.client.Client) -> bool:
    """
    Returns ``True`` if ``client`` has a non-malignant episode earlier than a
    malignant episode; ``False`` otherwise.

    :param client: A classified client with more than one episode
    """

    if client.status != omidb.client.Status.M:
        return False

    nm_date = None  # earliest non-malignant date
    m_date = None  # most recent malignant date

    for episode in client.episodes:
        if None in (episode.id, episode.opened_date):
            continue

        date = episode.opened_date

        if episode.has_malignant_opinions:
            if (m_date is None) or (date > m_date):
                m_date = date
        else:
            if (nm_date is None) or (date < nm_date):
                nm_date = date

        if nm_date and m_date:
            if nm_date < m_date:
                return True

    return False


def filter_studies_by_event_type(
    client: omidb.client.Client, event_type: List[Event], exact_match: bool = True
) -> omidb.client.Client:
    """
    Remove studies (``omidb.study.Study``) from ``client`` whose event type
    includes one or all (if ``exact_match`` is ``True``) of those listed by
    ``event_type``.

    :param client: A classified client with more than one episode
    :param event_type: A list of event types
    :param exact_match: If ``True`` the event types of a study must all of
        those in ``event_type``.
    """

    client2 = copy.deepcopy(client)

    for idx, episode in enumerate(client.episodes):
        if not episode.studies:
            continue
        for study in episode.studies:
            if study.event_type:
                if exact_match and set(study.event_type) != set(event_type):
                    client2.episodes[idx].studies.remove(study)  # type: ignore
                elif not any([_ in study.event_type for _ in event_type]):
                    client2.episodes[idx].studies.remove(study)  # type: ignore

    return client2


def _earliest_date(episode: Episode) -> datetime.date:
    """
    Returns the earliest date from this episodes event dates.

    If no event dates are present, the earliest date is taken as
    ``Episode.diagnosis_date`` for ``Type.CI`` episodes and
    ``Episode.opened_date`` otherwise.
    Raises a ``ValueError`` if no dates are found.
    """

    dates: List[datetime.date] = []
    if episode.events is not None:
        for field in dataclasses.fields(episode.events):
            event = getattr(episode.events, field.name)
            if not event:
                continue
            if isinstance(event, list):
                for e in event:
                    if e.dates is not None:
                        dates += e.dates
            else:
                dates += event.dates if event.dates is not None else []
    if dates:
        return min(dates)
    if episode.type == Type.CI:
        if episode.diagnosis_date is not None:
            return episode.diagnosis_date  # type: ignore
    elif episode.opened_date is not None:
        return episode.opened_date  # type: ignore
    raise ValueError("No valid dates to compare")


def _find_next_episodes(
    episode: Episode,
    all_episodes: List[Episode],
) -> List[Episode]:
    all_episodes = sorted(all_episodes, key=_earliest_date)
    idx = all_episodes.index(episode)
    if idx < len(all_episodes) - 1:
        return all_episodes[idx + 1 :]
    return []


def _find_next_episode(
    episode: Episode,
    all_episodes: List[Episode],
) -> Optional[Episode]:
    all_episodes = sorted(all_episodes, key=_earliest_date)
    idx = all_episodes.index(episode)
    if idx < len(all_episodes) - 1:
        return all_episodes[idx + 1]
    return None


def _within_time_period(e1: Episode, e2: Episode, num_months: int) -> Optional[bool]:
    t1 = _earliest_date(e1)
    t2 = _earliest_date(e2)
    if t2 - t1 <= datetime.timedelta(num_months * 365 / 12):
        return True
    return False


def _is_time_bounded_cancer_prior(
    e1: Episode, e2: Episode, num_months: int = 39
) -> bool:
    test_time = _within_time_period(e1, e2, num_months)
    if test_time and e2.has_malignant_opinions:
        return True
    return False


def _is_time_bounded_ci_prior(
    e1: Episode, e2: Episode, num_months: int = 36
) -> Optional[bool]:
    """
    Returns `True` if the `Episode.is_interval_cancer` evaluates to `True` in
    the episode prior to `episode` and the prior is within 3 years of
    `episode`.
    """
    test_time = _within_time_period(e1, e2, num_months)
    if test_time and e2.type == Type.CI:
        if e2.is_interval_cancer:
            return True
        else:
            return None  # Ambiguous
    return False


def _is_not_cancer(ep: Episode) -> bool:
    if ep.status == Status.CI:
        return False
    if ep.type == Type.CI:
        return False
    if ep.status == Status.M:
        return False
    return True


def is_post_op(episode: Episode, all_episodes: List[Episode]) -> bool:
    """
    Returns `True` if *any* episode in `all_episodes` prior to `episode` is of
    type `Episode.CI` or includes surgery information.
    """
    all_episodes = sorted(all_episodes, key=_earliest_date)
    idx = all_episodes.index(episode)
    for ep in all_episodes[:idx]:
        if (
            ep.events is not None
            and ep.events.surgery is not None
            or ep.type == Type.CI
        ):
            return True
    return False


@enum.unique
class EpisodeOutcome(enum.Enum):
    """
    Summarises the overall outcome of an episode according to the event
    opinions and the episode type of this episode and subsequent episodes.
    """

    CI = "Interval Cancer"
    CIP = "Interval Cancer Prior"
    M = "Malignant"
    MP = "Malignant Prior"
    B = "Benign"
    NAB = "Normal with assessment and biopsy and subsequent episode"
    NA = "Normal with assessment and subsequent episode"
    N = "Normal with subsequent non-cancer episode"


@enum.unique
class UndefinedEpisodeOutcome(enum.Enum):
    """
    When an episode cannot be classified into one of `EpisodeOutcome`, the
    episode is undefined with a reason given by this enum.
    """

    InvalidEvents = "No events or invalidated"
    InvalidFollowUp = "Follow up too early"
    InvalidPrior = "Non-normal found, but CI/MP prior criteria not satisfied"
    InvalidCI = "Event opinions contradict CI episode type"
    DateError = "Episodes cannot be sorted"
    NoSubsequentEpisode = "Subsequent episode not found"


def episode_outcome(
    episode: Episode,
    all_episodes: List[Episode],
    num_months_ci_prior: Optional[int] = None,
    num_months_cancer_prior: Optional[int] = None,
    num_months_normal_follow_up: Optional[int] = None,
    num_months_benign_follow_up: Optional[int] = None,
) -> Tuple[Union[EpisodeOutcome, UndefinedEpisodeOutcome], Optional[str]]:
    """
    Classifies an episode into one of the enumerations defined by
    ``EpisodeOutcome``, or ``UndefinedEpisodeOutcome`` if the classification
    criteria is not satisfied based on logic and/or insufficient data. For
    episodes outcomes that depend directly on a subsequent episode, the ID of
    that episode is returned alongside the outcome.

    :param episode: The episode to classify
    :param all_episodes: All episodes belonging to a single client
    :param num_months_ci_prior: The maximum number of months between an
        interval cancer and *any* previous episode for a previous episode to be
        considered a prior (``EpisodeOutcome.CIP``). If ``None`` (the default) the time
        interval between adjacent episodes is not considered, and the proximal
        prior of the interval cancer qualifies as ``EpisodeOutcome.CIP``, regardless of
        the amount of time between them.
    :param num_months_cancer_prior: The maximum number of months between a
        malignant episode (not an interval cancer) and *any* previous episode for a
        previous episode to be considered a prior (``EpisodeOutcome.MP``). If ``None``
        (the default) the time interval between adjacent episodes is not considered, and
        the proximal prior of the malignant cancer qualifies as ``EpisodeOutcome.MP``,
        regardless of the amount of time between them.
    :param num_months_normal_follow_up: The number of months _after_ which a
        second normal episode must exist in order for ``episode`` to be classified
        as normal. If ``None``, no follow-up is required.
    :param num_months_benign_follow_up: The number of months _after_ which a
        second non-malignant episode must exist in order for ``episode`` to be
        classified as benign. If ``None``, no follow-up is required, which may be
        preferred if the NBSS biopsy opinion is to serve as the ground truth.

    The classification logic is as follows (statuses are presented in order or
    precedence):

    ``EpisodeOutcome.CI``:
        - Episode type is CI; and
        - Episode outcome is malignant or has no surgery/biopsy information
        - If the second condition is not satisfied, status is undefined (``None``)

    ``EpisodeOutcome.M``:
        - Episode outcome is malignant

    ``EpisodeOutcome.CIP``:
        - Episode is the proximal prior of a CI episode occurring within
          ``num_months_ci_prior`` months. If ``num_months_ci_prior`` is
          not ``None``, ``episode`` is considered a prior if it falls within
          the specified time interval.
        - The CI episode must have a malignant outcome or no surgery/biopsy
          information. Otherwise, the prior is undefined (``None``).

    ``EpisodeOutcome.MP``:
        - Episode is the proximal prior of a malignant episode occurring within
          ``num_months_cancer_prior`` months. If ``num_months_ci_prior`` is
          not ``None``, ``episode`` is considered a prior if it falls within
          the specified time interval.

    ``EpisodeOutcome.B``:
        - If ``num_months_benign_follow_up`` is not
          ``None``, there must be a non-cancer after the specified number of
          months, and no cancer before. If ``None`` (the default), there is no
          requirement for a subsequent episode.
        - Episode contains biopsy events, leading to an ``EpisodeStatus`` of ``B``.


    ``EpisodeOutcome.NAB``:
        -  If ``num_months_normal_follow_up`` is not
          ``None``, there must be a non-cancer after the specified number of
          months, and no cancer before. If ``None`` (the default), there is no
          requirement for a subsequent episode.
        - Episode contains assessment and biopsy events.

    ``EpisodeOutcome.NA``:
        - The subsequent episode is not ``EpisodeOutcome.CI`` or
          ``EpisodeOutcome.M``. If ``num_months_normal_follow_up`` is not
          ``None``, there must be a non-cancer after the specified number of
          months, and no cancer before.
        - Episode contains an assessment event.
        - Episode and subsequent episodes' events must be present (not ``None``)

    ``EpisodeOutcome.N``:
        - The subsequent episode is not ``EpisodeOutcome.CI`` or
          ``EpisodeOutcome.M``. If ``num_months_normal_follow_up`` is not
          ``None``, there must be a non-cancer after the specified number of
          months, and no cancer before.
        - Episode and subsequent episode events must be present (not ``None``)


    Notes:
    - For statuses that depend on the input episode and subsequent episode(s),
      these episodes must not have any empty (``None``) left/right opinions
      in any non-screening event (if any).

    - If the above conditions are not satisfied and no ``ValueError``s are
      thrown, the episode type undefined: an ``UndefinedEpisodeOutcome`` is
      returned.

    - With the exception of interval cancers, the ``Type`` of episode is not
      considered when classifying an episode. You should pre-filter the
      episodes by ``Type`` if you would like restrict the classification to
      specific episode types, or otherwise adapt this function as needed.

    See also: ``is_post_op``.
    """
    episode_status = episode.status
    # Cancers
    if episode_status == Status.CI:
        return EpisodeOutcome.CI, None
    if episode.type == Type.CI:
        return UndefinedEpisodeOutcome.InvalidEvents, None
    if episode_status == Status.M:
        return EpisodeOutcome.M, None

    if episode.events is None or not _validate_events(episode):
        return UndefinedEpisodeOutcome.InvalidEvents, None

    """
    The next group depend on next episode and validated events, which may
    raise exceptions
    """
    try:
        for n_ep in _find_next_episodes(episode, all_episodes):
            if not _validate_events(n_ep):
                return UndefinedEpisodeOutcome.InvalidEvents, None
    except ValueError:
        return UndefinedEpisodeOutcome.DateError, None

    next_episode = _find_next_episode(episode, all_episodes)

    """
    if no follow-up, then we may only classify as normal or benign
    so long as a follow-up is not required. We need to check for
    a subsequent episode because priors trump negatives
    """
    if next_episode is None:
        if num_months_benign_follow_up is None:
            if episode_status == Status.B:
                return EpisodeOutcome.B, None

        if num_months_normal_follow_up is None:
            return _classify_normal_or_raise(episode), None

        return UndefinedEpisodeOutcome.NoSubsequentEpisode, None

    if num_months_ci_prior is None:
        num_months_ci_prior = ceil(
            12 * (_earliest_date(next_episode) - _earliest_date(episode)).days / 365.0
        )

    if num_months_cancer_prior is None:
        num_months_cancer_prior = ceil(
            12 * (_earliest_date(next_episode) - _earliest_date(episode)).days / 365.0
        )

    longest_window = max(num_months_ci_prior or 0, num_months_cancer_prior or 0)

    for n_ep in _find_next_episodes(episode, all_episodes):
        if _within_time_period(episode, n_ep, longest_window):
            if n_ep.type == Type.CI:
                res = _classify_ci_prior(episode, n_ep, num_months_ci_prior)
                if res:
                    return EpisodeOutcome.CIP, n_ep.id
                elif res is None:
                    return UndefinedEpisodeOutcome.InvalidEvents, n_ep.id
                else:
                    return UndefinedEpisodeOutcome.InvalidPrior, n_ep.id
            elif n_ep.status == Status.M:
                res = _classify_m_prior(episode, n_ep, num_months_cancer_prior)
                if res:
                    return EpisodeOutcome.MP, n_ep.id
                else:
                    return UndefinedEpisodeOutcome.InvalidPrior, n_ep.id
        else:
            break

    """
    Here we check if a follow-up is required or not.
    If not, then early return benign or normal.
    Otherwise, continue using the approproate number of months
    as the minimum inter-episode time delta.
    """
    if episode_status == Status.B:
        if num_months_benign_follow_up is None:
            return EpisodeOutcome.B, None
        num_months = num_months_benign_follow_up
    else:
        if num_months_normal_follow_up is None:
            return _classify_normal_or_raise(episode), None
        num_months = num_months_normal_follow_up

    """
    A follow-up is required.
    `episode_to_check` is either the first
    episode AFTER the `num_months_normal_follow_up` or `None`.
    If `None` then we assume we have no valid follow-up.
    """
    episode_to_check = None

    for ep in _find_next_episodes(episode, all_episodes):
        if ep.events is None:
            return UndefinedEpisodeOutcome.InvalidEvents, ep.id
        # is there a cancer inside window?
        if _within_time_period(episode, ep, num_months):
            if not _is_not_cancer(ep):
                return UndefinedEpisodeOutcome.InvalidPrior, ep.id
        # first episode outside window
        else:
            episode_to_check = ep
            break

    if episode_to_check is None:
        return UndefinedEpisodeOutcome.InvalidFollowUp, None

    # first episode outside window is cancer, so must be invalid prior
    if not _is_not_cancer(episode_to_check):
        return UndefinedEpisodeOutcome.InvalidPrior, episode_to_check.id

    # At this point we have two non-malignant neighbouring episodes
    if episode_status == Status.B:
        return EpisodeOutcome.B, episode_to_check.id

    return _classify_normal_or_raise(episode), episode_to_check.id


def _classify_normal_or_raise(episode: Episode) -> EpisodeOutcome:
    normal = _classify_normal(episode.status)
    if normal is None:
        raise ValueError(
            f"Failed to classify episode {episode.id}, unexpected condition"
        )
    return normal


def _classify_normal(s: Optional[Status]) -> Optional[EpisodeOutcome]:
    if s == Status.NAB:
        return EpisodeOutcome.NAB
    if s == Status.NA:
        return EpisodeOutcome.NA
    if s == Status.N:
        return EpisodeOutcome.N
    return None


def _classify_prior(
    episode: Episode,
    next_episode: Episode,
    num_months_ci_prior: Optional[int],
    num_months_cancer_prior: Optional[int],
) -> Optional[bool]:
    if next_episode.type == Type.CI:
        return _classify_ci_prior(episode, next_episode, num_months_ci_prior)
    else:
        return _classify_m_prior(episode, next_episode, num_months_ci_prior)


def _classify_ci_prior(
    episode: Episode, next_episode: Episode, num_months_ci_prior: Optional[int]
) -> Optional[bool]:
    if next_episode.type == Type.CI:
        if num_months_ci_prior is not None:
            is_valid_ci_prior = _is_time_bounded_ci_prior(
                episode, next_episode, num_months_ci_prior
            )
        elif next_episode.is_interval_cancer:
            is_valid_ci_prior = True
        else:
            is_valid_ci_prior = None

        if is_valid_ci_prior:
            return True
        elif is_valid_ci_prior is None:
            return None
    return False


def _classify_m_prior(
    episode: Episode, next_episode: Episode, num_months_cancer_prior: Optional[int]
) -> bool:
    if (next_episode.type != Type.CI) and next_episode.has_malignant_opinions:
        if num_months_cancer_prior is None or _is_time_bounded_cancer_prior(
            episode, next_episode, num_months_cancer_prior
        ):
            return True
    return False



import omidb
from omidb import client_parser as cp

"""
Artificial data for testing purposes (assume properties have no relation)
"""

episode_data = {
    "LESION": {
        "R": {
            "1": {
                "CystAspirated": "N",
                "LesionDescription": "Mass with calcification",
                "LesionNotes": None,
                "LesionPosition": "RB3",
            }
        }
    },
    "BIOPSYWIDE": {
        "Opinion": None,
        "R": {
            "1": {
                "BenignLesions": None,
                "CalcificationOnSpecimen": None,
                "DatePerformed": "1950-02-22",
                "DcisGrade": "NDI",
                "DiagnosticSetOutcome": "W+",
                "DiseaseGrade": "G3",
                "EpithelialProliferation": None,
                "HER2ReceptorScore": None,
                "HER2ReceptorStatus": None,
                "HistologicalCalcification": "A",
                "HormoneERScore": None,
                "HormoneERStatus": None,
                "HormonePRScore": None,
                "HormonePRStatus": None,
                "InSituComponents": "NID",
                "InsituPresent": None,
                "InvasiveComponents": "IDC",
                "InvasivePresent": "Y",
                "InvasiveType": "IN",
                "InvasiveTypeOther": None,
                "LocalisationType": "U",
                "Location": None,
                "LymphNode": "N",
                "MalignancyType": "b",
                "NeedleSpecimenType": "WBN",
                "Opinion": "B5",
                "SideCode": "R",
                "SpecimenType": None,
                "left_opinion": "Normal",
                "right_opinion": "Malignant",
            }
        },
        "dateperformed": "1950-02-22",
        "diagnosticsetoutcome": "W+",
        "left_opinion": None,
        "right_opinion": "Malignant",
    },
    "SURGERY": {
        "Opinion": None,
        "R": {
            "1": {
                "AxillaryMetsType": "MET",
                "AxillaryNodesNumberPositive": "1",
                "AxillaryNodesPositive": None,
                "AxillarySpecimenType": None,
                "BenignLesions": None,
                "BiopsySpecimenType": "WX",
                "CalcificationsOnSpecimen": None,
                "CancerOnKC62": "N",
                "Chemotherapy": None,
                "DatePerformed": "1950-05-13",
                "DcisGrade": "NDI",
                "DcisGrowthPatterns": "GC GP",
                "DiagnosticProcedure": None,
                "DiseaseExtent": "DEL",
                "DiseaseGrade": "G3",
                "EpithelialProliferation": None,
                "Event": "PTR",
                "ExcisionMarginDistance": "2.0",
                "ExcisionMargins": "ED",
                "HER2ReceptorScore": "1+",
                "HER2ReceptorStatus": "RN",
                "HistologicalCalcification": "A",
                "HormoneERScore": "8",
                "HormoneERStatus": "RP",
                "HormonePRScore": "8",
                "HormonePRStatus": "RP",
                "InSituComponents": "NID",
                "InvasiveComponents": "IDC",
                "InvasiveType": "IN",
                "InvasiveTypeOther": None,
                "LocalTrialCode": None,
                "Microinvasion": "MNP",
                "NationalTrial": None,
                "NonsurgicalTreatments": None,
                "Opinion": "H5",
                "ReconstructiveProcedure": "NO",
                "SizeDuctalOnly": None,
                "TherapeuticProcedures": "SD AC",
                "TreatmentProcedure": "WLE",
                "WholeSizeOfTumour": "16.0",
                "left_opinion": "Normal",
                "right_opinion": "Malignant",
            }
        },
        "dateperformed": "1950-05-13",
        "diagnosticsetoutcome": "H+",
        "left_opinion": None,
        "right_opinion": "Malignant",
    },
}


def test_parse_lesion() -> None:

    parser = cp.ClientParser("", {}, {}, [], False)
    parsed_lesions = parser._parse_lesions(episode_data)

    for side, side_data in episode_data["LESION"].items():
        for id, lesion_data in side_data.items():
            lesion = parsed_lesions[id]
            assert lesion.id == id
            assert lesion.side.name == side
            assert lesion.description.value == lesion_data["LesionDescription"]
            assert lesion.position.name == lesion_data["LesionPosition"]


def test_parse_biopsy_wide_lesion() -> None:

    parser = cp.ClientParser("", {}, {}, [], False)

    for side, side_data in episode_data["LESION"].items():
        for id, _ in side_data.items():

            lesion_event = parser._parse_biopsy_wide_lesion(
                episode_data, side, id)
            event_data = episode_data["BIOPSYWIDE"][side][id]
            assert str(lesion_event.date) == event_data["DatePerformed"]
            assert str(
                lesion_event.disease_grade.name) == event_data["DiseaseGrade"]
            assert (
                str(lesion_event.invasive_components[0].name)
                == event_data["InvasiveComponents"]
            )
            assert str(
                lesion_event.invasive_type.name) == event_data["InvasiveType"]
            assert (
                str(lesion_event.insitu_components[0].name)
                == event_data["InSituComponents"]
            )
            assert str(lesion_event.dcis_grade.name) == event_data["DcisGrade"]
            assert str(
                lesion_event.malignant_type.name) == event_data["MalignancyType"]
            assert lesion_event.opinion.name == event_data.get("Opinion")


def test_parse_surgery_lesion() -> None:

    parser = cp.ClientParser("", {}, {}, [], False)

    for side, side_data in episode_data["LESION"].items():
        for id, _ in side_data.items():

            lesion_event = parser._parse_surgery_lesion(episode_data, side, id)

            event_data = episode_data["SURGERY"][side][id]
            assert str(lesion_event.date) == event_data["DatePerformed"]
            assert str(
                lesion_event.disease_grade.name) == event_data["DiseaseGrade"]
            assert (
                str(lesion_event.invasive_components[0].name)
                == event_data["InvasiveComponents"]
            )
            assert str(
                lesion_event.invasive_type.name) == event_data["InvasiveType"]
            assert (
                str(lesion_event.insitu_components[0].name)
                == event_data["InSituComponents"]
            )
            assert str(lesion_event.dcis_grade.name) == event_data["DcisGrade"]
            assert lesion_event.opinion.name == event_data.get("Opinion")



from .parser import DB  # noqa
from . import (  # noqa
    image,
    mark,
    series,
    study,
    events,
    episode,
    client,
    utilities,
    filters,
    classificationtools,
    commands,
)
from loguru import logger

logger.disable("omidb")



from dataclasses import dataclass
from typing import Optional, Set
import enum


@enum.unique
class BenignClassification(enum.Enum):
    """Description of the benign feature"""

    coarse = "coarse_or_popcorn-like"
    egg_shell = "egg_shell_or_rim"
    lucent_centred = "lucent-centred"
    rod = "rod-like"
    round_puncate = "round_and_punctate"


@enum.unique
class Conspicuity(enum.Enum):
    """Indicates this marks conspicuity"""

    obvious = "Obvious"
    subtle = "Subtle"
    very_subtle = "Very_subtle"
    occult = "Occult"
    not_recalled = "not_recalled"


@enum.unique
class MassClassification(enum.Enum):
    """Description of the mass border"""

    ill_defined = "ill_defined"
    spiculated = "spiculated"
    well_defined = "well_defined"
    other = "other"
    unknown = "unknown"
    lucent_centred = "lucent-centred"
    course = "coarse_or_popcorn-like"
    egg_shell = "egg_shell_or_rim"


@dataclass
class BoundingBox:
    """2D coordinates defining the mark"""

    x1: int
    y1: int
    x2: int
    y2: int


@dataclass
class Mark:
    """
    A ground-truth region of interest made by an expert radiologist. A mark
    includes the region of interest and descriptions characterising the
    lesion.

    Note that boolean attributes default to ``None``, rather than ``False``, to
    reduce misinterpretation when the source attribute has not been set by the
    Radiologist. Interpret as 'no value'.

    :param id: An identifier for the annotation
    :param boundingBox: 2D coordinates defining the lower-left (x1, y1) and
        upper-right (x2, y2) corners of the mark
    :param conspicuity: Enumeration indicating mark conspicuity
    :param lesion_ids: An set of identifiers for the lesions associated with
        this mark; unique within episode only. Corresponds to the
        `LinkedNBSSLesionNumber` field of the ImageDB JSON, which itself
        maps to the key of the lesion referenced by the NBSS JSON.
    :param architectural_distortion: The mark highlights architectural
        distortion
    :param dystrophic_calcification: The mark highlights dystrophic
        calcification
    :param fat_necrosis: The mark highlights an area of fat necrosis
    :param focal_asymmetry: Indicates focal asymmetry
    :param mass: The marked lesion is a mass
    :param suspicious_calcifications: Indicates a suspicious calcification
        cluster
    :param milk_of_calcium: The mark highlights milk of calcium mammographic
        features
    :param other_benign_cluster: The mark surrounds a benign calcification
        cluster not described by other parameters
    :param plasma_cell_mastitis: The marked area indicates plasma-cell-mastitis
    :param benign_skin_feature: The mark captures a benign skin lesion,
        calcified or other
    :param calcifications: The marked area has calcifications
    :param suture_calcification: The marked area indicates suture
        calcifications
    :param vascular_feature: The marked area indicates vascular calcifications
    :param benign_classification: Enumeration describing the benign feature
    :param mass_classification: Enumeration describing the mass border
    """

    id: str
    boundingBox: BoundingBox
    conspicuity: Conspicuity
    lesion_ids: Optional[Set[str]] = None
    architectural_distortion: Optional[bool] = None
    dystrophic_calcification: Optional[bool] = None
    fat_necrosis: Optional[bool] = None
    focal_asymmetry: Optional[bool] = None
    mass: Optional[bool] = None
    suspicious_calcifications: Optional[bool] = None
    milk_of_calcium: Optional[bool] = None
    other_benign_cluster: Optional[bool] = None
    plasma_cell_mastitis: Optional[bool] = None
    benign_skin_feature: Optional[bool] = None
    calcifications: Optional[bool] = None
    suture_calcification: Optional[bool] = None
    vascular_feature: Optional[bool] = None
    benign_classification: Optional[BenignClassification] = None
    mass_classification: Optional[MassClassification] = None



import pydicom
import matplotlib
import matplotlib.pyplot as plt
from dataclasses import dataclass
from typing import List, Optional, Dict, Any, Callable
from .mark import Mark


@dataclass
class LoaderParams:
    client_id: str
    study_id: str
    series_id: str
    image_id: str


DicomLoaderFunc = Callable[[LoaderParams], pydicom.dataset.FileDataset]
JsonLoaderFunc = Callable[[LoaderParams], Dict[str, Any]]


@dataclass
class DicomLoader:
    args: LoaderParams
    func: DicomLoaderFunc


@dataclass
class JsonLoader:
    args: LoaderParams
    func: JsonLoaderFunc


@dataclass
class Image:
    """
    Container for a mammogram, stored in the DICOM_ format. An image can have
    zero or more marks for one or many lesions.

    :param id: SOP Instance UID, a unique identifier
    :param dcm_path: Path to the dicom image
    :param json_path: Path to the JSON file storing DICOM metadata
    :param marks: A list of marks or annotations, represented by
        :class:`omidb.mark.Mark`

    .. _DICOM: https://www.dicomstandard.org/
    """

    id: str
    dcm_loader: Optional[DicomLoader] = None
    json_loader: Optional[JsonLoader] = None
    marks: Optional[List[Mark]] = None
    _dcm: Optional[pydicom.FileDataset] = None
    _json: Optional[Dict[str, Any]] = None

    @property
    def dcm(self) -> Optional[pydicom.FileDataset]:
        """
        Returns a :class:`pydicom.dataset.FileDataset`, representing a parsed DICOM file
        """
        if not self._dcm and self.dcm_loader is not None:
            self._dcm = self.dcm_loader.func(self.dcm_loader.args)
        return self._dcm

    @property
    def attributes(self) -> Optional[Dict[str, Any]]:
        """
        Access DICOM metadata via the JSON representation
        """

        if not self._json and self.json_loader is not None:
            self._json = self.json_loader.func(self.json_loader.args)
        return self._json

    def plot(
        self, ax: Optional[matplotlib.axes.Axes] = None
    ) -> Optional[matplotlib.image.AxesImage]:
        """
        Plot the dicom
        """

        if self.dcm is None:
            return None

        if not ax:
            fig, ax = plt.subplots()

        return ax.imshow(self.dcm.pixel_array, cmap=plt.cm.bone)



import enum
import omidb


def test_lookup():
    class E(enum.Enum):
        A = "a"
        B = "b"

    for e in E:
        got = omidb.utilities.enum_lookup(e.value, E)
        expected = getattr(E, e.name)
        assert got == expected

    for e in E:
        got = omidb.utilities.enum_lookup(e.name, E)
        expected = getattr(E, e.name)
        assert got == expected

    assert omidb.utilities.enum_lookup(None, E) is None


def test_nbss_str_to_enum_for_list_like():
    codes = (
        "IDC",
        "ILC",
        "IMC",
        "IMU",
        "IPX",
        "ITC",
        "Invalid"
    )

    value = (" ").join(codes)

    result = omidb.utilities.nbss_str_to_enum(
        value,
        omidb.lesion.InvasiveCarcinomaComponent,
        "InvasiveComponents",
    )
    expected = [
        omidb.lesion.InvasiveCarcinomaComponent.IDC,
        omidb.lesion.InvasiveCarcinomaComponent.ILC,
        omidb.lesion.InvasiveCarcinomaComponent.IMC,
        omidb.lesion.InvasiveCarcinomaComponent.IMU,
        omidb.lesion.InvasiveCarcinomaComponent.IPX,
        omidb.lesion.InvasiveCarcinomaComponent.ITC,
    ]

    assert result == expected

def test_nbss_str_to_enum_for_single_str():

    result = omidb.utilities.nbss_str_to_enum(
        "H2",
        omidb.events.Opinion,
    )

    assert result == omidb.events.Opinion.H2


import datetime
from dataclasses import dataclass, fields
from typing import Optional, List, Dict
import enum
from .events import Events, SideOpinion
from .study import Study
from .lesion import Lesion


@enum.unique
class Type(enum.Enum):
    """Type of episode"""

    CA = "Continued Assessment"
    CD = "Delayed Treatment"
    CF = "Follow-up after treatment"
    CI = "Interval case"
    CR = "Local Recurrence"
    F = "First Call"
    G = "GP Referral"
    H = "Higher Risk"
    N = "Non-rout Recall"
    R = "Routine Recall"
    S = "Self Referral"
    X = "Other"


@enum.unique
class Action(enum.Enum):
    """Episode action"""

    EC = "Early Recall for Clinic"
    ES = "Early Recall for Screening"
    FN = "Fine Needle Aspiration"
    FP = "Follow-up (Post-treatment)"
    FV = "Further X-ray views"
    IP = "Inpatient biopsy"
    MT = "Medical Treatment"
    NA = "No Action from this procedure"
    R2 = "Routine second film opinion (obsolete)"
    RC = "Review in clinic"
    RF = "Referral to consultant/GP"
    RR = "Routine recall for screening"
    ST = "Surgical Treatment"
    TR = "Repeat Film (technical)"
    WB = "Wide Bore Needle"


@enum.unique
class Status(enum.Enum):
    """
    Summarises the status of an episode according to the event opinions
    and the episode type. Does not consider the properties of any other episode.
    """

    CI = "Interval Cancer"
    M = "Malignant"
    B = "Benign"
    NAB = "Normal with assessment and biopsy"
    NA = "Normal with assessment"
    N = "Normal"


@dataclass
class Episode:
    """
    An episode contains a set of medical procedures or events associated with
    the treatment or diagnosis of a clinical condition. Medical imaging studies
    are included with each episode, and are typically (but not always) linked
    to one or more events.

    :param id: NBSS episode identifier, only unique within client
    :param events: A set of medical procedures associated with the episode
    :param studies: List of :class:`omidb.study.Study` s where collections of
        screening and diagnostic images reside
    :param type: Enumeration defining the type of episode
    :param action: Enumeration defining the action outcome of the episode
    :param opened_date: Date that the episode opened
    :param closed_date: Date that the episode closed
    :param diagnosis_date: Date that the interval cancer (if applicable) was diagnosed
    :param is_closed: boolean signifying whether the episode is closed
    :param lesions: A list of :class:`omidb.lesion.Lesion` s, examined in the
        episode
    :param actual_opened_year: True year that the episode was opened.
    """

    id: str
    events: Optional[Events] = None
    studies: Optional[List[Study]] = None
    type: Optional[Type] = None
    action: Optional[Action] = None
    opened_date: Optional[datetime.date] = None
    closed_date: Optional[datetime.date] = None
    diagnosis_date: Optional[datetime.date] = None
    is_closed: Optional[bool] = None
    lesions: Optional[Dict[str, Lesion]] = None
    actual_opened_year: Optional[int] = None

    @property
    def has_benign_opinions(self) -> bool:
        """
        Returns ``True`` if the container ``e.events`` contains a
        :class:`omidb.events.Events.surgery` or
        :class:`omidb.events.Events.biopsy_wide` or
        :class:`omidb.events.Events.biopsy_fine` or where the left or right
        opinion is benign (:class:`omidb.events.SideOpinion.OB`)
        """

        if self.events is None:
            return False
        for event in (
            self.events.surgery,
            self.events.biopsy_wide,
            self.events.biopsy_fine,
        ):
            if event is not None and (
                event.left_opinion == SideOpinion.OB
                or event.right_opinion == SideOpinion.OB
            ):
                return True

        return False

    @property
    def has_malignant_opinions(self) -> bool:
        """
        Returns ``True`` if the container ``e.events`` contains a
        :class:`omidb.events.Events.surgery` or
        :class:`omidb.events.Events.biopsy_wide` or
        :class:`omidb.events.Events.biopsy_fine` or where the left or right
        opinion is malignant (:class:`omidb.events.SideOpinion.OM`)
        """

        if self.events is None:
            return False

        for event in (
            self.events.surgery,
            self.events.biopsy_wide,
            self.events.biopsy_fine,
        ):
            if event is not None and (
                event.left_opinion == SideOpinion.OM
                or event.right_opinion == SideOpinion.OM
            ):
                return True

        return False

    @property
    def is_interval_cancer(self) -> bool:
        """
        Return ``True`` if:
            ``episode.type`` is ``omidb.episode.Type.CI``
            ``episode.has_malignant_opinions`` evaluates to ``True``
        Return ``False`` otherwise.
        """

        if self.type != Type.CI:
            return False

        if self.has_malignant_opinions:
            return True

        if self.events is not None:
            for event in (
                self.events.surgery,
                self.events.biopsy_wide,
                self.events.biopsy_fine,
            ):
                if event is not None:
                    return False
        return True

    @property
    def status(self) -> Optional[Status]:
        if self.is_interval_cancer:
            return Status.CI
        if self.type == Type.CI:
            return None
        if self.has_malignant_opinions:
            return Status.M
        if self.has_benign_opinions:
            return Status.B
        if (
            self.events is None
            or not _validate_events(self)
            or self.events.surgery is not None
        ):
            return None
        if self.events.assessment is not None:
            if (
                self.events.biopsy_fine is not None
                or self.events.biopsy_wide is not None
            ):
                return Status.NAB
            else:
                return Status.NA
        return Status.N


def _validate_events(e: Episode) -> bool:
    """
    If any non-screening event contains null opinions, returns ``False``.

    This is to protect against any 'null' events that imply the patient
    underwent more than just a screening examination and the outcome is
    unknown.
    """
    if e.events is None:
        return True
    for field in fields(e.events):
        event = getattr(e.events, field.name)
        if event is None:
            continue

        if field.name == "screening":
            continue

        if event.left_opinion is None and event.right_opinion is None:
            return False
    return True



import pytest
import datetime
from omidb import episode, events
from omidb.classificationtools import _earliest_date, _is_time_bounded_ci_prior


def test_earliest_date_event_date_over_episode_opened() -> None:
    expected = datetime.date(2000, 1, 1)
    ep = episode.Episode(
        id="1",
        events=events.Events(screening=[events.Screening(dates=[expected])]),
        opened_date=expected - datetime.timedelta(days=1),
    )
    assert _earliest_date(ep) == expected


def test_earliest_date_episode_opened_no_events() -> None:
    expected = datetime.date(2000, 1, 1)
    ep = episode.Episode(
        id="1",
        events=events.Events(screening=[events.Screening()]),
        opened_date=expected,
    )
    assert _earliest_date(ep) == expected


def test_earliest_date_interval_cancer() -> None:
    expected = datetime.date(2000, 1, 1)
    ep = episode.Episode(
        id="1",
        events=events.Events(screening=[events.Screening()]),
        opened_date=expected - datetime.timedelta(days=1),  # ignored!
        diagnosis_date=expected,
        type=episode.Type.CI,
    )
    assert _earliest_date(ep) == expected


def test_earliest_date_multiple_screens() -> None:
    expected = datetime.date(2000, 1, 1)
    ep = episode.Episode(
        id="1",
        events=events.Events(
            screening=[events.Screening(dates=[expected, datetime.date(2000, 2, 1)])],
        ),
    )
    assert _earliest_date(ep) == expected


def test_earliest_date_multiple_events() -> None:
    expected = datetime.date(2000, 1, 1)
    ep = episode.Episode(
        id="1",
        events=events.Events(
            screening=[events.Screening(dates=[expected + datetime.timedelta(days=1)])],
            surgery=events.BaseEvent(dates=[expected]),
        ),
    )
    assert _earliest_date(ep) == expected


def test_earliest_date_episode_raises_when_no_dates() -> None:
    ep = episode.Episode(
        id="1",
        events=events.Events(screening=[events.Screening()]),
    )
    with pytest.raises(ValueError):
        _earliest_date(ep)


def test_earliest_date_episode_raises_when_ci_has_only_opened_date() -> None:
    ep = episode.Episode(
        id="1",
        events=events.Events(screening=[events.Screening()]),
        opened_date=datetime.date(2000, 1, 1),
        type=episode.Type.CI,
    )
    with pytest.raises(ValueError):
        _earliest_date(ep)



import pytest
import omidb
from collections import namedtuple
import tempfile
import pathlib
from typing import Iterator


mark_data = {
    "ArchitecturalDistortion": "ArchitecturalDistortion",
    "BenignClassification": "coarse_or_popcorn-like",
    "Conspicuity": "Subtle",
    "Dystrophic": "Dystrophic",
    "FatNecrosis": "FatNecrosis",
    "FocalAsymmetry": "FocalAsymmetry",
    "Height": "240",
    "LesionID": 6711,
    "LinkedNBSSLesionNumber": "1,2",
    "MarkID": 10241,
    "Mass": 1,
    "MassClassification": "spiculated",
    "MilkOfCalcium": "MilkOfCalcium",
    "OtherBenignCluster": "OtherBenignCluster",
    "PlasmaCellMastitis": "PlasmaCellMastitis",
    "Skin": "Skin",
    "SuspiciousCalcifications": "SuspiciousCalcifications",
    "SutureCalcification": "SutureCalcification",
    "Vascular": "Vascular",
    "Width": "327",
    "WithCalcification": "WithCalcification",
    "X1": "756",
    "X2": "1183",
    "Y1": "1487",
    "Y2": "1733",
}


Dirs = namedtuple("Dirs", "root omidb data images")


@pytest.fixture
def dirs() -> Iterator[Dirs]:

    with tempfile.TemporaryDirectory() as root_dir:
        root = pathlib.Path(root_dir)
        omidb_dir = root / "omidb"
        data_dir = omidb_dir / "data"
        images_dir = omidb_dir / "images"

        omidb_dir.mkdir()
        data_dir.mkdir()
        images_dir.mkdir()

        yield Dirs(root, omidb_dir, data_dir, images_dir)


def test_parse_mark(dirs: Dirs) -> None:

    parser = omidb.client_parser.ClientParser("", {}, {}, [], False)

    mark = parser._parse_mark(mark_data)

    assert mark.architectural_distortion
    assert mark.benign_classification == omidb.mark.BenignClassification.coarse
    assert mark.conspicuity == omidb.mark.Conspicuity.subtle
    assert mark.dystrophic_calcification
    assert mark.fat_necrosis
    assert mark.focal_asymmetry
    assert mark.id == str(mark_data["MarkID"])
    assert mark.lesion_ids == set(["1", "2"])
    assert mark.milk_of_calcium
    assert mark.other_benign_cluster
    assert mark.plasma_cell_mastitis
    assert mark.benign_skin_feature
    assert mark.suspicious_calcifications
    assert mark.suture_calcification
    assert mark.vascular_feature
    assert mark.calcifications

    assert mark.boundingBox.x1 == int(mark_data.get("X1", 0))
    assert mark.boundingBox.x2 == int(mark_data.get("X2", 0))
    assert mark.boundingBox.y1 == int(mark_data.get("Y1", 0))
    assert mark.boundingBox.y2 == int(mark_data.get("Y2", 0))



from dataclasses import dataclass
import enum
from typing import List
from . import episode


@enum.unique
class Status(enum.Enum):
    """
    Summarises the overall status of a client according to the event opinions
    within an episode, and the type of episode itself.
    """

    B = "Benign"
    M = "Malignant"
    N = "Normal"
    CI = "Interval Cancer"


@dataclass
class Client:
    """
    A client represents a patient who has attended the NHS screening programme.

    Each client will have one or more :class:`omidb.episode.Episode` s, for
    which (anonymised and pseudonymised) NBSS information can be found.

    :param id: Client identifier, typically a four letter ID followed by a
        series of digits, e.g. `optm1`, `demd7050`
    :param episodes: A list of :class:`omidb.episode.Episode` s associated with
        the client
    :param site: code denoting the clinical site/centre where screening took
        place
    """

    id: str
    episodes: List[episode.Episode]
    site: str

    @property
    def status(self) -> Status:
        """
        Determines the :class:`omidb.client.Status` by first classifying
        each episode according to its type and the surgery and biopsy opinions,
        and subsequently applying the following precedence rules
        (from top to bottom: highest to lowest precedence):

        - :class:`Status.CI`: if any interval cancer episode

        - :class:`Status.M`: if any malignant episode (based on surgery/biopsy opinions)

        - :class:`Status.B`: if any benign episode (based on surgery/biopsy opinions)

        - :class:`Status.N`: otherwise
        """

        for ep in self.episodes:
            if ep.type == episode.Type.CI:
                return Status.CI

        for ep in self.episodes:
            if ep.has_malignant_opinions:
                return Status.M

        for ep in self.episodes:
            if ep.has_benign_opinions:
                return Status.B

        return Status.N



import datetime
import dataclasses
from typing import List, Dict, Optional, Any
from loguru import logger
from . import utilities
from .client import Client
from .episode import Episode
from . import episode
from . import image as im
from .study import Study
from .series import Series
from .lesion import (
    Lesion,
    LesionDescription,
    LesionAssessment,
    LesionSurgery,
    LesionClinical,
    LesionIntervalCancer,
    LesionBiopsyWide,
    LesionBiopsyFine,
    LesionPosition,
    Side,
    InvasiveCarcinomaComponent,
    HistologicalGrade,
    InSituCarcinomaComponent,
    InvasiveCarcinomaType,
    DCISGrade,
    MalignancyType,
)
from .mark import (
    BenignClassification,
    Conspicuity,
    MassClassification,
    BoundingBox,
    Mark,
)

from .events import (
    Event,
    Events,
    BreastScreeningData,
    Screening,
    BaseEvent,
    Opinion,
    SideOpinion,
)


class ClientParser:
    def __init__(
        self,
        id: str,
        nbss: Dict[str, Any],
        imagedb: Dict[str, Any],
        studies: Optional[List[str]] = None,
        distinct_event_study_links: bool = True,
        json_loader: Optional[im.JsonLoaderFunc] = None,
        dcm_loader: Optional[im.DicomLoaderFunc] = None,
    ):
        self.id = id
        self.nbss = nbss
        self.imagedb = imagedb
        self.studies = studies
        self.distinct_event_study_links = distinct_event_study_links
        self.json_loader = json_loader
        self.dcm_loader = dcm_loader
        self._episode_id: Optional[str] = None

    def __call__(self) -> Client:
        episodes = self.parse_episodes()
        return Client(id=self.id, episodes=episodes, site=self.imagedb["Site"])

    def logmsg(self, msg: str) -> str:
        _msg = f"{self.id}"
        if self._episode_id is not None:
            _msg = f"{_msg}/Episode {self._episode_id}"
        _msg = f"{_msg}: {msg}"
        return _msg

    def parse_episodes(self) -> List[Episode]:
        events: Dict[str, Optional[Events]] = {}
        for episode_id, nbss_episode in self.nbss.items():
            if isinstance(nbss_episode, dict):
                events[episode_id] = self.parse_events(nbss_episode)

        episode_studies = self.studies_by_episode(events)

        out = []

        for episode_id, episode_events in events.items():
            self._episode_id = episode_id

            this_episodes_studies = episode_studies.get(episode_id, None)

            if not this_episodes_studies:
                logger.warning(self.logmsg("Episode in NBSS but not in IMAGEDB"))
            elif self.distinct_event_study_links and episode_events is not None:
                self.ensure_distinct_study_event_links(
                    this_episodes_studies, episode_events
                )

            nbss_episode = self.nbss[episode_id]
            is_closed = nbss_episode.get("EpisodeIsClosed") == "Y"

            ep_type: Optional[episode.Type] = utilities.nbss_str_to_enum(
                nbss_episode.get("EpisodeType"), episode.Type
            )

            ep_action: Optional[episode.Action] = utilities.nbss_str_to_enum(
                nbss_episode.get("EpisodeAction"), episode.Action
            )

            opened_date = utilities.date_or_none(nbss_episode, "EpisodeOpenedDate")
            closed_date = utilities.date_or_none(nbss_episode, "EpisodeClosedDate")
            diagnosis_date = utilities.date_or_none(
                nbss_episode, "IntervalCancerDateOfDiagnosis"
            )

            actual_opened_year = (
                int(nbss_episode.get("ActualEpisodeOpenedYear"))
                if nbss_episode.get("ActualEpisodeOpenedYear", None)
                else None
            )

            lesions = self._parse_lesions(nbss_episode)

            out.append(
                Episode(
                    id=episode_id,
                    events=episode_events,
                    studies=this_episodes_studies,
                    type=ep_type,
                    action=ep_action,
                    is_closed=is_closed,
                    lesions=lesions,
                    actual_opened_year=actual_opened_year,
                    opened_date=opened_date,
                    closed_date=closed_date,
                    diagnosis_date=diagnosis_date,
                )
            )

            self._episode_id = None
        return out

    def studies_by_episode(
        self,
        events: Dict[str, Optional[Events]],
    ) -> Dict[str, List[Study]]:
        # Studies for each episode
        episode_studies: Dict[str, List[Study]] = {}

        if not self.has_studies():
            return episode_studies

        for study_id, study_data in self.imagedb["STUDIES"].items():
            logger.info(
                self.logmsg(f"Attempting to link study {study_id} to an episode")
            )

            if self.studies is not None and study_id not in self.studies:
                logger.info(
                    self.logmsg(
                        f"{study_id} found in ImageDB, but not "
                        "in provided list of studies, skipping"
                    )
                )
                continue

            study_date = utilities.date_or_none(study_data, "StudyDate")

            if study_date is None:
                logger.warning(self.logmsg(f"{study_id} has no study date"))
            # Try to extract episode ID by study-date <->event-date
            elif (
                ("EpisodeID" not in study_data)
                or (not study_data["EpisodeID"])
                or (study_data["EpisodeID"] not in events)
            ):
                logger.warning(
                    self.logmsg(
                        f"Episode {study_data.get('EpisodeID')} (in IMAGEDB) not found"
                        f" in NBSS for study {study_id}, attempting link via event"
                        " dates (will replace episode ID)"
                    )
                )

                episode_id = self.find_episode_id_by_event_dates(study_date, events)

                if episode_id is not None:
                    study_data["EpisodeID"] = episode_id

            # If stil not episode ID, skip
            if ("EpisodeID" not in study_data) or (not study_data["EpisodeID"]):
                logger.error(f"EpisodeID not found for study {study_id}, skipping")

                continue

            # if event and (study_date in event_dates): fallback to episode id
            matched_events: List[Event] = []
            if study_date is not None and study_data["EpisodeID"] in events:
                matched_events = self.match_events(
                    events[study_data["EpisodeID"]],
                    study_date,
                )
            else:
                logger.warning(
                    self.logmsg(
                        f"Episode {study_data['EpisodeID']} (IMAGEDB) has no events"
                        "(episode not found in NBSS)"
                    )
                )
                continue

            # Now add studies to the episode
            study = Study(
                id=study_id,
                series=self.parse_series(study_id, study_data),
                date=study_date,
                event_type=matched_events,
            )

            if study_data["EpisodeID"] not in episode_studies:
                episode_studies[study_data["EpisodeID"]] = [study]
            else:
                episode_studies[study_data["EpisodeID"]].append(study)

        return episode_studies

    def parse_series(
        self,
        study_iuid: str,
        study_data: Dict[str, Any],
    ) -> List[Series]:
        series_list = []

        for series, series_dic in study_data.items():
            if not (isinstance(series_dic, dict) and utilities.IUID_REG.match(series)):
                continue

            image_list = [
                im for im in series_dic.keys() if utilities.IUID_REG.match(im)
            ]

            images = []
            for image in image_list:
                marks: List[Mark] = []

                image_marks_data = series_dic.get(image)
                if isinstance(image_marks_data, dict):
                    for _, mark_data in image_marks_data.items():
                        try:
                            marks.append(self._parse_mark(mark_data))
                        except Exception:
                            logger.exception(
                                f"Failed to parse mark data of {self.id} {image}.dcm"
                            )
                            continue

                args = im.LoaderParams(self.id, study_iuid, series, image)

                if self.dcm_loader is not None:
                    dcm_loader = im.DicomLoader(args, self.dcm_loader)
                else:
                    dcm_loader = None

                if self.json_loader is not None:
                    json_loader = im.JsonLoader(args, self.json_loader)
                else:
                    json_loader = None

                images.append(
                    im.Image(
                        id=image,
                        dcm_loader=dcm_loader,
                        json_loader=json_loader,
                        marks=marks,
                    )
                )

            series_list.append(Series(id=series, images=images))
        return series_list

    def has_studies(self) -> bool:
        if ("STUDIES" not in self.imagedb) or (
            not isinstance(self.imagedb["STUDIES"], dict)
        ):
            logger.error("No studies listed in IMAGEDB for client {self.id]}")
            return False
        return True

    def match_events(
        self,
        episode_events: Optional[Events],
        study_date: datetime.date,
    ) -> List[Event]:
        matched_events: List[Event] = []
        if episode_events is None:
            return matched_events

        for field in dataclasses.fields(episode_events):
            event_list = getattr(episode_events, field.name)
            if event_list is None:
                continue

            if not isinstance(event_list, list):
                event_list = [event_list]

            for e in event_list:
                if study_date in e.dates:
                    matched_events.append(getattr(Event, field.name))

        if not matched_events:
            logger.warning(self.logmsg(f"No events match study date {study_date}"))

            for field in dataclasses.fields(episode_events):
                event = getattr(episode_events, field.name)
                if event is None:
                    continue
                matched_events.append(getattr(Event, field.name))
                logger.warning(
                    self.logmsg(
                        f"Linked {field.name} event to study by episode ID only"
                    )
                )

        if len(matched_events) > 1:
            logger.warning(self.logmsg("Multiple events linked"))
            if self.distinct_event_study_links:
                logger.info(self.logmsg("Dropping matched events as not distinct"))
                matched_events = []
        return matched_events

    def find_episode_id_by_event_dates(
        self, study_date: datetime.date, events: Dict[str, Any]
    ) -> Optional[str]:
        for episode_id, episode_events in events.items():
            if episode_events is None:
                continue
            for field in dataclasses.fields(episode_events):
                event = getattr(episode_events, field.name)
                if event is None:
                    continue

                if isinstance(event, list):
                    event_dates = [date for e in event for date in e.dates]
                else:
                    event_dates = event.dates

                if event and (study_date in event_dates):
                    logger.info(
                        self.logmsg(f"Linked study date {study_date} to {episode_id}")
                    )
                    return episode_id
        return None

    def ensure_distinct_study_event_links(
        self, studies: List[Study], events: Optional[Events]
    ) -> None:
        """
        Filters (mutates) the `event_type` property of an episode's `studies` such
        that studies-event links are distinct.
        """

        for idx, study1 in enumerate(studies):
            for study2 in studies[idx:]:
                if study1 == study2:
                    continue

                # Two studies have the same date, drop any links
                if study1.date == study2.date:
                    logger.warning(
                        self.logmsg(
                            "Dropping events linked to "
                            f"{study1.id} and {study2.id} as same date "
                        )
                    )
                    study1.event_type = []
                    study2.event_type = []
                    continue

                # Two studies have the different dates,
                # but linked to same event
                if study1.event_type and study1.event_type == study2.event_type:
                    event_list = getattr(events, study1.event_type[0].name)

                    if not isinstance(event_list, list):
                        event_list = [event_list]

                    # Pool the dates if multiple screens...
                    event_dates = [date for e in event_list for date in e.dates]

                    if study1.date not in event_dates:
                        logger.warning(
                            self.logmsg(f"Dropping events linked to {study1.id}")
                        )

                        study1.event_type = []

                    if study2.date not in event_dates:
                        logger.warning(
                            self.logmsg(f"Dropping events linked to {study2.id}")
                        )
                        study2.event_type = []

    def parse_events(self, episode_data: Dict[str, Any]) -> Optional[Events]:
        event_kwargs = {
            "screening": episode_data.get("SCREENING", None),
            "assessment": episode_data.get("ASSESSMENT", None),
            "biopsy_wide": episode_data.get("BIOPSYWIDE", None),
            "biopsy_fine": episode_data.get("BIOPSYFINE", None),
            "clinical": episode_data.get("CLINICAL", None),
            "surgery": episode_data.get("SURGERY", None),
        }

        has_events = False
        for key, value in event_kwargs.items():
            if value is None:
                continue

            has_events = True
            # SCREENING must exist in order to parse OTHERSCREENING
            if key == "screening":
                event_kwargs["screening"] = [self.parse_screening_event(value)]

                other_screening = episode_data.get("OTHERSCREENING", None)
                if other_screening is not None:
                    for _, screen in other_screening.items():
                        event_kwargs["screening"].append(
                            self.parse_screening_event(screen)
                        )
            else:
                event_kwargs[key] = self.parse_base_event(value)

        if has_events:
            return Events(**event_kwargs)
        return None

    def parse_base_event(self, data: Dict[str, Any]) -> BaseEvent:
        dates: List[datetime.date] = []

        for _, side in enumerate(["L", "R"]):
            side_data = data.get(side)

            if not side_data:
                continue

            for _, lesion in side_data.items():
                if lesion.get("DatePerformed"):
                    date = utilities.date_or_none(lesion, "DatePerformed")
                    if date is not None and date not in dates:
                        dates.append(date)

        # Some events may not have lesions due to null ID
        # so add higher-level dateperformed to improve linking
        date = utilities.date_or_none(data, "dateperformed")
        if date is not None and date not in dates:
            dates.append(date)

        left_opinion = utilities.nbss_str_to_enum(data.get("left_opinion"), SideOpinion)

        right_opinion = utilities.nbss_str_to_enum(
            data.get("right_opinion"), SideOpinion
        )

        return BaseEvent(
            left_opinion=left_opinion, right_opinion=right_opinion, dates=dates
        )

    def parse_screening_event(self, data: Dict[str, Any]) -> Screening:
        breast_data: List[Optional[BreastScreeningData]] = [None, None]
        dates: List[datetime.date] = []

        for i, side in enumerate(["L", "R"]):
            side_data = data.get(side)

            if not side_data:
                continue

            date = utilities.date_or_none(side_data, "DateTaken")
            if date is not None and date not in dates:
                dates.append(date)

            opinion: Optional[Opinion] = utilities.nbss_str_to_enum(
                side_data.get("Opinion"), Opinion
            )

            breast_data[i] = BreastScreeningData(
                date=date,
                equipment_make_model=side_data.get("EquipmentMakeModel"),
                opinion=opinion,
            )

        left_opinion: Optional[SideOpinion] = utilities.nbss_str_to_enum(
            data.get("left_opinion"), SideOpinion
        )

        right_opinion: Optional[SideOpinion] = utilities.nbss_str_to_enum(
            data.get("right_opinion"), SideOpinion
        )

        return Screening(
            left_opinion=left_opinion,
            right_opinion=right_opinion,
            left=breast_data[0],
            right=breast_data[1],
            dates=dates,
        )

    def _parse_lesions(
        self, episode_data: Dict[str, Any]
    ) -> Optional[Dict[str, Lesion]]:
        lesions_data = episode_data.get("LESION")
        lesions: Dict[str, Lesion] = {}

        if lesions_data is None:
            return None

        for side in ["L", "R"]:
            side_data = lesions_data.get(side)
            if side_data is None:
                continue

            for lesion_id, lesion_data in side_data.items():
                lesion = Lesion(
                    id=lesion_id,
                    side=utilities.nbss_str_to_enum(side, Side),
                    cyst_aspirated=lesion_data.get("CystAspirated", False) == "Y",
                    description=utilities.nbss_str_to_enum(
                        lesion_data.get("LesionDescription"), LesionDescription
                    ),
                    position=utilities.nbss_str_to_enum(
                        lesion_data.get("LesionPosition"), LesionPosition
                    ),
                    notes=lesion_data.get("LesionNotes"),
                    assessment=self._parse_assessment_lesion(
                        episode_data, side, lesion_id
                    ),
                    surgery=self._parse_surgery_lesion(episode_data, side, lesion_id),
                    clinical=self._parse_clinical_lesion(episode_data, side, lesion_id),
                    biopsy_wide=self._parse_biopsy_wide_lesion(
                        episode_data, side, lesion_id
                    ),
                    biopsy_fine=self._parse_biopsy_fine_lesion(
                        episode_data, side, lesion_id
                    ),
                    interval_cancer=self._parse_interval_cancer_lesion(
                        episode_data, side, lesion_id
                    ),
                )

                lesions[lesion_id] = lesion

        return lesions

    def _parse_assessment_lesion(
        self,
        episode_data: Dict[str, Any],
        side: str,
        lesion_id: str,
    ) -> Optional[LesionAssessment]:
        try:
            data = episode_data["ASSESSMENT"][side][lesion_id]
        except KeyError:
            return None

        date = utilities.date_or_none(data, "DatePerformed")
        return LesionAssessment(date=date)

    def _parse_surgery_lesion(
        self,
        episode_data: Dict[str, Any],
        side: str,
        lesion_id: str,
    ) -> Optional[LesionSurgery]:
        try:
            data = episode_data["SURGERY"][side][lesion_id]
        except Exception:
            return None

        date = utilities.date_or_none(data, "DatePerformed")

        ivcmps: Optional[List[InvasiveCarcinomaComponent]] = utilities.nbss_str_to_enum(
            data.get("InvasiveComponents"),
            InvasiveCarcinomaComponent,
            "InvasiveComponents",
        )

        iscmps: Optional[List[InSituCarcinomaComponent]] = utilities.nbss_str_to_enum(
            data.get("InSituComponents"),
            InSituCarcinomaComponent,
            "InSituComponents",
        )

        return LesionSurgery(
            date=date,
            invasive_components=ivcmps,
            disease_grade=utilities.nbss_str_to_enum(
                data.get("DiseaseGrade"), HistologicalGrade
            ),
            invasive_type=utilities.nbss_str_to_enum(
                data.get("InvasiveType"),
                InvasiveCarcinomaType,
            ),
            insitu_components=iscmps,
            dcis_grade=utilities.nbss_str_to_enum(data.get("DcisGrade"), DCISGrade),
            opinion=utilities.nbss_str_to_enum(data.get("Opinion"), Opinion),
        )

    def _parse_clinical_lesion(
        self,
        episode_data: Dict[str, Any],
        side: str,
        lesion_id: str,
    ) -> Optional[LesionClinical]:
        try:
            data = episode_data["CLINICAL"][side][lesion_id]
        except KeyError:
            return None

        return LesionClinical(
            date=utilities.date_or_none(data, "DatePerformed"),
            opinion=utilities.nbss_str_to_enum(data.get("Opinion"), Opinion),
        )

    def _parse_biopsy_fine_lesion(
        self,
        episode_data: Dict[str, Any],
        side: str,
        lesion_id: str,
    ) -> Optional[LesionBiopsyFine]:
        try:
            data = episode_data["BIOPSYFINE"][side][lesion_id]
        except KeyError:
            return None

        return LesionBiopsyFine(
            date=utilities.date_or_none(data, "DatePerformed"),
            opinion=utilities.nbss_str_to_enum(data.get("Opinion"), Opinion),
        )

    def _parse_biopsy_wide_lesion(
        self,
        episode_data: Dict[str, Any],
        side: str,
        lesion_id: str,
    ) -> Optional[LesionBiopsyWide]:
        try:
            data = episode_data["BIOPSYWIDE"][side][lesion_id]
        except KeyError:
            return None

        ivcmps: Optional[List[InvasiveCarcinomaComponent]] = utilities.nbss_str_to_enum(
            data.get("InvasiveComponents"),
            InvasiveCarcinomaComponent,
            "InvasiveComponents",
        )

        iscmps: Optional[List[InSituCarcinomaComponent]] = utilities.nbss_str_to_enum(
            data.get("InSituComponents"),
            InSituCarcinomaComponent,
            "InSituComponents",
        )

        return LesionBiopsyWide(
            date=utilities.date_or_none(data, "DatePerformed"),
            invasive_components=ivcmps,
            disease_grade=utilities.nbss_str_to_enum(
                data.get("DiseaseGrade"), HistologicalGrade
            ),
            invasive_type=utilities.nbss_str_to_enum(
                data.get("InvasiveType"), InvasiveCarcinomaType
            ),
            insitu_components=iscmps,
            dcis_grade=utilities.nbss_str_to_enum(data.get("DcisGrade"), DCISGrade),
            malignant_type=utilities.nbss_str_to_enum(
                data.get("MalignancyType"),
                MalignancyType,
            ),
            opinion=utilities.nbss_str_to_enum(data.get("Opinion"), Opinion),
        )

    def _parse_interval_cancer_lesion(
        self,
        episode_data: Dict[str, Any],
        side: str,
        lesion_id: str,
    ) -> Optional[LesionIntervalCancer]:
        try:
            data = episode_data["INTERVALCANCER"][side][lesion_id]
        except KeyError:
            return None

        return LesionIntervalCancer(date=utilities.date_or_none(data, "DatePerformed"))

    def _parse_mark(self, mark_data: Dict[str, Any]) -> Mark:
        args: Dict[str, Any] = {}

        for param_name, key in zip(
            (
                "architectural_distortion",
                "dystrophic_calcification",
                "fat_necrosis",
                "focal_asymmetry",
                "mass",
                "suspicious_calcifications",
                "milk_of_calcium",
                "other_benign_cluster",
                "plasma_cell_mastitis",
                "benign_skin_feature",
                "calcifications",
                "suture_calcification",
                "vascular_feature",
            ),
            (
                "ArchitecturalDistortion",
                "Dystrophic",
                "FatNecrosis",
                "FocalAsymmetry",
                "Mass",
                "SuspiciousCalcifications",
                "MilkOfCalcium",
                "OtherBenignCluster",
                "PlasmaCellMastitis",
                "Skin",
                "WithCalcification",
                "SutureCalcification",
                "Vascular",
            ),
        ):
            args[param_name] = True if mark_data.get(key, None) else None

        args["benign_classification"] = utilities.nbss_str_to_enum(
            mark_data.get("BenignClassification"), BenignClassification
        )

        args["conspicuity"] = utilities.nbss_str_to_enum(
            mark_data.get("Conspicuity"), Conspicuity
        )

        args["mass_classification"] = utilities.nbss_str_to_enum(
            mark_data.get("MassClassification"), MassClassification
        )

        try:
            ids = mark_data["LinkedNBSSLesionNumber"].split(",")
            args["lesion_ids"] = set([str(int(_)) for _ in ids])
        except Exception:
            args["lesion_ids"] = None

        args["id"] = str(mark_data["MarkID"])

        args["boundingBox"] = BoundingBox(
            x1=int(mark_data["X1"]),
            y1=int(mark_data["Y1"]),
            x2=int(mark_data["X2"]),
            y2=int(mark_data["Y2"]),
        )

        return Mark(**args)



import omidb


def test_lesion_is_invasive_no_attributes() -> None:

    lesion = omidb.lesion.Lesion("a", omidb.lesion.Side.L)

    assert not lesion.is_invasive


def test_lesion_is_invasive_malignancy_types() -> None:

    for mtype, expected in zip(
        (
            omidb.lesion.MalignancyType.a,
            omidb.lesion.MalignancyType.b,
            omidb.lesion.MalignancyType.c,
        ),
        (False, True, False),
    ):

        biopsy_wide = omidb.lesion.LesionBiopsyWide(malignant_type=mtype)

        lesion = omidb.lesion.Lesion("", omidb.lesion.Side.L, biopsy_wide=biopsy_wide)

        assert lesion.is_invasive == expected


def test_lesion_is_invasive_biopsy_conditions() -> None:

    biopsy_wides = [
        omidb.lesion.LesionBiopsyWide(
            invasive_components=omidb.lesion.InvasiveCarcinomaComponent.IDC
        ),
        omidb.lesion.LesionBiopsyWide(
            invasive_type=omidb.lesion.InvasiveCarcinomaType.IN
        ),
        omidb.lesion.LesionBiopsyWide(disease_grade=omidb.lesion.HistologicalGrade.G1),
    ]

    for biopsy_wide in biopsy_wides:
        lesion = omidb.lesion.Lesion("a", omidb.lesion.Side.L, biopsy_wide=biopsy_wide)

        assert lesion.is_invasive


def test_lesion_is_invasive_surgery_conditions() -> None:

    surgeries = [
        omidb.lesion.LesionSurgery(
            invasive_components=omidb.lesion.InvasiveCarcinomaComponent.IDC
        ),
        omidb.lesion.LesionSurgery(invasive_type=omidb.lesion.InvasiveCarcinomaType.IN),
        omidb.lesion.LesionSurgery(disease_grade=omidb.lesion.HistologicalGrade.G1),
    ]

    for surgery in surgeries:
        lesion = omidb.lesion.Lesion("a", omidb.lesion.Side.L, surgery=surgery)

        assert lesion.is_invasive


def test_lesion_is_not_insitu_if_invasive(mocker) -> None:  # type: ignore

    mocker.patch("omidb.lesion.Lesion.is_invasive", new_callable=lambda: True)

    # This condition would otherwise result in `is_insitu` being True
    biopsy_wide = omidb.lesion.LesionBiopsyWide(
        malignant_type=omidb.lesion.MalignancyType.a
    )

    lesion = omidb.lesion.Lesion("", omidb.lesion.Side.L, biopsy_wide=biopsy_wide)

    assert not lesion.is_insitu


def test_lesion_is_insitu_malignancy_types() -> None:

    for mtype, expected in zip(
        (
            omidb.lesion.MalignancyType.a,
            omidb.lesion.MalignancyType.b,
            omidb.lesion.MalignancyType.c,
        ),
        (True, False, False),
    ):

        biopsy_wide = omidb.lesion.LesionBiopsyWide(malignant_type=mtype)

        lesion = omidb.lesion.Lesion("", omidb.lesion.Side.L, biopsy_wide=biopsy_wide)

        assert lesion.is_insitu == expected


def test_lesion_is_insitu_biopsy_conditions() -> None:

    biopsy_wides = [
        omidb.lesion.LesionBiopsyWide(
            insitu_components=omidb.lesion.InSituCarcinomaComponent.NID
        ),
        omidb.lesion.LesionBiopsyWide(dcis_grade=omidb.lesion.DCISGrade.NDH),
    ]

    for biopsy_wide in biopsy_wides:
        lesion = omidb.lesion.Lesion("a", omidb.lesion.Side.L, biopsy_wide=biopsy_wide)

        assert lesion.is_insitu


def test_lesion_is_insitu_surgery_conditions() -> None:

    surgeries = [
        omidb.lesion.LesionSurgery(
            insitu_components=omidb.lesion.InSituCarcinomaComponent.NID
        ),
        omidb.lesion.LesionSurgery(dcis_grade=omidb.lesion.DCISGrade.NDH),
    ]

    for surgery in surgeries:
        lesion = omidb.lesion.Lesion("a", omidb.lesion.Side.L, surgery=surgery)

        assert lesion.is_insitu


def test_lesion_grade_none() -> None:
    lesion = omidb.lesion.Lesion("a", omidb.lesion.Side.L)
    assert lesion.grade == None


def test_lesion_grade() -> None:

    params = [
        [
            {
                "disease_grade": omidb.lesion.HistologicalGrade.G1,
                "dcis_grade": omidb.lesion.DCISGrade.NDH,
            },
            {
                "disease_grade": omidb.lesion.HistologicalGrade.G2,
                "dcis_grade": omidb.lesion.DCISGrade.NDL,
            },
            omidb.lesion.HistologicalGrade.G2,
        ],
        [
            {
                "disease_grade": omidb.lesion.HistologicalGrade.G1,
                "dcis_grade": omidb.lesion.DCISGrade.NDH,
            },
            {"dcis_grade": omidb.lesion.DCISGrade.NDL},
            omidb.lesion.HistologicalGrade.G1,
        ],
        [
            {"dcis_grade": omidb.lesion.DCISGrade.NDH},
            {"dcis_grade": omidb.lesion.DCISGrade.NDL},
            omidb.lesion.DCISGrade.NDL,
        ],
        [
            {"dcis_grade": omidb.lesion.DCISGrade.NDH},
            {"insitu_components": None},
            omidb.lesion.DCISGrade.NDH,
        ],
    ]

    for i in range(len(params)):

        b = omidb.lesion.LesionBiopsyWide(**params[i][0])

        s = omidb.lesion.LesionSurgery(**params[i][1])

        lesion = omidb.lesion.Lesion("", omidb.lesion.Side.L, biopsy_wide=b, surgery=s)

        assert lesion.grade == params[i][2]


def test_lesion_status_none() -> None:
    lesion = omidb.lesion.Lesion("a", omidb.lesion.Side.L)
    assert lesion.status is None


def test_lesion_status_invasive(mocker) -> None:
    mocker.patch("omidb.lesion.Lesion.is_invasive", new_callable=lambda: True)
    mocker.patch("omidb.lesion.Lesion.is_insitu", new_callable=lambda: False)
    lesion = omidb.lesion.Lesion("a", omidb.lesion.Side.L)
    assert lesion.status == omidb.lesion.Status.Invasive


def test_lesion_status_invasive_trumps(mocker) -> None:
    mocker.patch("omidb.lesion.Lesion.is_invasive", new_callable=lambda: True)
    mocker.patch("omidb.lesion.Lesion.is_insitu", new_callable=lambda: True)
    lesion = omidb.lesion.Lesion("a", omidb.lesion.Side.L)
    assert lesion.status == omidb.lesion.Status.Invasive


def test_lesion_status_insitu(mocker) -> None:
    mocker.patch("omidb.lesion.Lesion.is_invasive", new_callable=lambda: False)
    mocker.patch("omidb.lesion.Lesion.is_insitu", new_callable=lambda: True)
    lesion = omidb.lesion.Lesion("a", omidb.lesion.Side.L)
    assert lesion.status == omidb.lesion.Status.Insitu



import datetime
import itertools
from typing import List, Optional, Union
from dataclasses import dataclass
from .events import Opinion
import enum


Side = enum.Enum("Side", "L R")

Status = enum.Enum("Status", "Invasive Insitu")


@enum.unique
class LesionDescription(enum.Enum):
    AS = "Asymmetry"
    CA = "Calcification only"
    CY = "Cyst"
    DS = "Distortion"
    LN = "Lymph node"
    MA = "Mass"
    MC = "Mass with calcification"
    NA = "No significant abnormality"
    ZZ = "Clinical abnormality"


def _lesion_positions() -> List[str]:
    result = ["LM", "RM"]
    for pos in itertools.product("LR", "ABCDE", "1234"):
        result.append("".join(pos))
    return result


LesionPosition = enum.Enum("LesionPosition", _lesion_positions())  # type: ignore


@enum.unique
class LocalisationType(enum.Enum):
    S = "Skin Marker"
    U = "Ultrasound Guidance"
    X = "X-Ray Guidance"


@enum.unique
class SiteIndicator(enum.Enum):
    M = "Multiple"
    S = "Single"


@enum.unique
class InvasiveCarcinomaComponent(enum.Enum):
    IDC = "Ductal/NST"
    ILC = "Lobular"
    IMC = "Medullary like"
    IMU = "Mucinous"
    IPX = "Other component"
    ITC = "Tubular/cribriform"


@enum.unique
class HistologicalGrade(enum.Enum):
    G1 = "I"
    G2 = "II"
    G3 = "III"
    NA = "Not assessable"


@enum.unique
class InvasiveCarcinomaType(enum.Enum):
    IN = "Ductal/NST"
    IP = "Pure Special Type IM = Mixed"
    IO = "Other"


@enum.unique
class InSituCarcinomaComponent(enum.Enum):
    NID = "Ductal"
    NIL = "Lobular"
    NIP = "Paget’s"


@enum.unique
class DCISGrade(enum.Enum):
    NDH = "High"
    NDI = "Intermediate"
    NDL = "Low"
    NDN = "Not assessable"


@enum.unique
class MalignancyType(enum.Enum):
    a = "In-situ"
    b = "Invasive"
    c = "Not assessable"


@enum.unique
class CancerType(enum.Enum):
    D = "Ductal"
    L = "Lobular"
    S = "Special"


@dataclass
class LesionAssessment:
    date: Optional[datetime.date] = None


@dataclass
class LesionSurgery:
    date: Optional[datetime.date] = None
    invasive_components: Optional[List[InvasiveCarcinomaComponent]] = None
    disease_grade: Optional[HistologicalGrade] = None
    invasive_type: Optional[InvasiveCarcinomaType] = None
    insitu_components: Optional[List[InSituCarcinomaComponent]] = None
    dcis_grade: Optional[DCISGrade] = None
    opinion: Optional[Opinion] = None


@dataclass
class LesionBiopsyWide:
    date: Optional[datetime.date] = None
    invasive_components: Optional[List[InvasiveCarcinomaComponent]] = None
    disease_grade: Optional[HistologicalGrade] = None
    invasive_type: Optional[InvasiveCarcinomaType] = None
    insitu_components: Optional[List[InSituCarcinomaComponent]] = None
    dcis_grade: Optional[DCISGrade] = None
    malignant_type: Optional[MalignancyType] = None
    opinion: Optional[Opinion] = None


@dataclass
class LesionBiopsyFine:
    date: Optional[datetime.date] = None
    opinion: Optional[Opinion] = None


@dataclass
class LesionIntervalCancer:
    date: Optional[datetime.date] = None


@dataclass
class LesionClinical:
    date: Optional[datetime.date] = None
    opinion: Optional[Opinion] = None


@dataclass
class Lesion:
    id: str
    side: Optional[Side] = None
    cyst_aspirated: Optional[bool] = None
    description: Optional[LesionDescription] = None
    position: Optional[LesionPosition] = None
    notes: Optional[str] = None
    # localisation_needed: Optional[bool] = None  # not available
    # localisation_type: Optional[LocalisationType] = None  # not available
    # site_indicator: Optional[SiteIndicator] = None  # not available
    assessment: Optional[LesionAssessment] = None
    clinical: Optional[LesionClinical] = None
    surgery: Optional[LesionSurgery] = None
    biopsy_wide: Optional[LesionBiopsyWide] = None
    biopsy_fine: Optional[LesionBiopsyFine] = None
    interval_cancer: Optional[LesionIntervalCancer] = None

    @property
    def status(self) -> Optional[Status]:
        if self.is_invasive:
            return Status.Invasive
        elif self.is_insitu:
            return Status.Insitu
        else:
            return None

    @property
    def is_invasive(self) -> bool:
        if self.biopsy_wide:
            if self.biopsy_wide.malignant_type == MalignancyType.b:
                return True

            if (
                self.biopsy_wide.invasive_components
                or self.biopsy_wide.invasive_type
                or self.biopsy_wide.disease_grade  # TODO: What if NA?
            ):
                return True

        if self.surgery and (
            self.surgery.invasive_components
            or self.surgery.invasive_type
            or self.surgery.disease_grade
        ):  # TODO: What if NA?
            return True

        return False

    @property
    def is_insitu(self) -> bool:
        if self.is_invasive:
            return False

        if self.biopsy_wide:
            if self.biopsy_wide.malignant_type == MalignancyType.a:
                return True

            # This condition can probably be dropped, but kept for consistency
            if self.biopsy_wide.malignant_type and (
                self.biopsy_wide.malignant_type != MalignancyType.c
            ):
                return False

            if self.biopsy_wide.insitu_components or self.biopsy_wide.dcis_grade:
                return True

        if self.surgery and (self.surgery.insitu_components or self.surgery.dcis_grade):
            return True

        return False

    @property
    def grade(self) -> Optional[Union[HistologicalGrade, DCISGrade]]:
        if self.surgery and self.surgery.disease_grade:
            return self.surgery.disease_grade

        if self.biopsy_wide and self.biopsy_wide.disease_grade:
            return self.biopsy_wide.disease_grade

        if self.surgery and self.surgery.dcis_grade:
            return self.surgery.dcis_grade

        if self.biopsy_wide and self.biopsy_wide.dcis_grade:
            return self.biopsy_wide.dcis_grade

        return None



from typing import List
from dataclasses import dataclass
import matplotlib
import matplotlib.pyplot as plt
from .image import Image


@dataclass
class Series:
    """
    A collection of images taken during one examination by one modality, for a
    given position of the patient on the acquisition device. In Full-Field
    Digital Mammography, each image is typically associated with one unique
    series.

    :param id: Series Instance UID, a unique identifier
    :param images: A list of :class:`Image` objects for the examination
    """

    id: str
    images: List[Image]

    @property
    def num_images(self) -> int:
        """
        The number of images in this series
        """
        return len(self.images)

    def plot(self) -> matplotlib.pyplot.Axes:
        """
        Convenience method for plotting all images in the series
        """

        if self.num_images == 1:
            return self.images[0].plot()

        fig, axes = plt.subplots(1, self.num_images)

        for i, image in enumerate(self.images):
            image.plot(axes[i])

        return axes



import omidb


def get_ci_ep() -> omidb.episode.Episode:
    return omidb.episode.Episode(
        "a", studies=None, events=None, type=omidb.episode.Type.CI
    )


def get_benign_ep() -> omidb.episode.Episode:

    event = omidb.events.BaseEvent(left_opinion=omidb.events.SideOpinion.OB)

    events = omidb.events.Events(surgery=event)
    return omidb.episode.Episode("a", events, None)  # type: ignore


def get_malignant_ep() -> omidb.episode.Episode:

    event = omidb.events.BaseEvent(left_opinion=omidb.events.SideOpinion.OM)

    events = omidb.events.Events(surgery=event)
    return omidb.episode.Episode("a", events, None)  # type: ignore


def get_ep() -> omidb.episode.Episode:

    event = omidb.events.BaseEvent(left_opinion=omidb.events.SideOpinion.ON)

    events = omidb.events.Events(assessment=event)
    return omidb.episode.Episode("a", events, None)  # type: ignore


def test_classify_client() -> None:

    n_test = ([get_ep()], omidb.client.Status.N)

    b_test = ([get_ep(), get_benign_ep()], omidb.client.Status.B)

    m_test = ([get_ep(), get_benign_ep(), get_malignant_ep()], omidb.client.Status.M)

    ci_test = (
        [get_ep(), get_benign_ep(), get_malignant_ep(), get_ci_ep()],
        omidb.client.Status.CI,
    )

    for episodes, expected in [n_test, b_test, m_test, ci_test]:

        client = omidb.client.Client(
            "demd1", site=None, episodes=episodes  # type:ignore
        )

        assert client.status == expected



import pytest
import datetime
from omidb import episode, events
from omidb.classificationtools import _earliest_date, _is_time_bounded_ci_prior


def test_normals() -> None:
    episodes = [
        episode.Episode(
            id="1",
            events=events.Events(
                screening=[events.Screening(dates=[datetime.date(2000, 1, 1)])],
                assessment=events.BaseEvent(
                    left_opinion=events.SideOpinion.ON,
                ),
                biopsy_fine=events.BaseEvent(
                    left_opinion=events.SideOpinion.ON,
                ),
            ),
        ),
        episode.Episode(
            id="2",
            events=events.Events(
                screening=[events.Screening(dates=[datetime.date(2000, 2, 1)])],
                assessment=events.BaseEvent(
                    left_opinion=events.SideOpinion.ON,
                ),
            ),
        ),
        episode.Episode(
            id="3",
            events=events.Events(
                screening=[events.Screening(dates=[datetime.date(2000, 3, 1)])],
            ),
        ),
        episode.Episode(  # Null date but treated as valid
            id="4",
            events=events.Events(
                screening=[events.Screening()],
            ),
            opened_date=datetime.date(2000, 4, 1),
        ),
    ]

    expected = [
        episode.Status.NAB,
        episode.Status.NA,
        episode.Status.N,
        episode.Status.N,
    ]
    for ep, exp in zip(episodes, expected):
        assert ep.status == exp


def test_surgery_no_opinion() -> None:
    ep = episode.Episode(
        id="1",
        events=events.Events(
            surgery=events.BaseEvent(left_opinion=events.SideOpinion.ON),
        ),
    )

    assert ep.status is None


def test_invalid_events() -> None:
    ep = episode.Episode(  # Null date but treated as valid
        id="1",
        events=events.Events(clinical=events.BaseEvent()),
        opened_date=datetime.date(2000, 4, 1),
    )

    assert ep.status is None


def test_ci() -> None:
    episodes = [
        episode.Episode(
            id="1",
            type=episode.Type.CI,
        ),
        episode.Episode(
            id="2",
            events=events.Events(
                surgery=events.BaseEvent(left_opinion=events.SideOpinion.ON),
            ),
            type=episode.Type.CI,
        ),
        episode.Episode(
            id="2",
            events=events.Events(
                surgery=events.BaseEvent(left_opinion=events.SideOpinion.OM),
            ),
            type=episode.Type.CI,
        ),
    ]

    expected = [
        episode.Status.CI,
        None,
        episode.Status.CI,
    ]
    for ep, exp in zip(episodes, expected):
        assert ep.status == exp


def test_malignant() -> None:
    episodes = [
        episode.Episode(
            id="1",
            events=events.Events(
                surgery=events.BaseEvent(left_opinion=events.SideOpinion.OM),
            ),
            type=episode.Type.R,
        ),
        episode.Episode(
            id="2",
            events=events.Events(
                biopsy_fine=events.BaseEvent(left_opinion=events.SideOpinion.OM),
            ),
            type=episode.Type.R,
        ),
        episode.Episode(
            id="3",
            events=events.Events(
                biopsy_wide=events.BaseEvent(left_opinion=events.SideOpinion.OM),
            ),
        ),
    ]

    expected = [
        episode.Status.M,
        episode.Status.M,
        episode.Status.M,
    ]
    for ep, exp in zip(episodes, expected):
        assert ep.status == exp


def test_benign() -> None:
    episodes = [
        episode.Episode(
            id="1",
            events=events.Events(
                surgery=events.BaseEvent(left_opinion=events.SideOpinion.OB),
            ),
            type=episode.Type.R,
        ),
        episode.Episode(
            id="2",
            events=events.Events(
                biopsy_fine=events.BaseEvent(left_opinion=events.SideOpinion.OB),
            ),
            type=episode.Type.R,
        ),
        episode.Episode(
            id="3",
            events=events.Events(
                biopsy_wide=events.BaseEvent(left_opinion=events.SideOpinion.OB),
            ),
        ),
    ]

    expected = [
        episode.Status.B,
        episode.Status.B,
        episode.Status.B,
    ]
    for ep, exp in zip(episodes, expected):
        assert ep.status == exp



import enum
from dataclasses import dataclass
import datetime
from typing import List, Optional


@enum.unique
class Opinion(enum.Enum):
    """Opinion after a procedure"""

    A1 = "Assess normal"
    A2 = "Assess benign"
    A3 = "Assess uncert'n"
    A4 = "Assess suspic"
    A5 = "Assess malig"
    B1 = "Unsatis/Normal"
    B2 = "Benign"
    B3 = "Benign unc mal"
    B4 = "Susp of malig"
    B5 = "Malignant"
    BA = "Clinical"
    C1 = "Cyt unsatis"
    C2 = "Cyt benign"
    C3 = "Cyt atypia"
    C4 = "Cyt susp malig"
    C5 = "Cyt malig"
    H0 = "Hist unreported"
    H1 = "Histol normal"
    H2 = "Histol benign"
    H5 = "Hist malignant"
    I1 = "Imaging normal"
    I2 = "Imaging benign"
    I3 = "Imaging uncertain"
    I4 = "Imaging suspicious"
    I5 = "Imaging malig"
    MRI1 = "MRI normal"
    MRI2 = "MRI benign"
    MRI3 = "MRI indeterminate"
    MRI4 = "MRI suspicious"
    MRI5 = "MRI malig"
    OB = "O Benign"
    OM = "O Malignant"
    ON = "O Normal"
    OS = "O Suspicious"
    OU = "O Uncertain"
    P1 = "Clin normal"
    P2 = "Clin benign"
    P3 = "Clin uncertain"
    P4 = "Clin suspicious"
    P5 = "Clin malignant"
    R1 = "Rad normal"
    R2 = "Rad benign"
    R3 = "Rad uncertain"
    R4 = "Rad suspicious"
    R5 = "Rad malig"
    RB = "R Benign"
    RM = "R Malignant"
    RN = "R Normal"
    RO = "R Unreported"
    RS = "R Suspicious"
    RU = "R Uncertain"
    SH = "History suspic"
    U1 = "USS normal"
    U2 = "USS benign"
    U3 = "USS uncertain"
    U4 = "USS suspicious"
    U5 = "USS malig"


@enum.unique
class SideOpinion(enum.Enum):
    """Side-specific opinion"""

    OB = "Benign"
    OM = "Malignant"
    ON = "Normal"
    OS = "Suspicious"
    OU = "Uncertain"


@enum.unique
class Event(enum.Enum):
    """Type of event. Corresponds to the fields of :class:`omidb.events.Events`"""

    assessment = "assessment"
    biopsy_fine = "biopsy_fine"
    biopsy_wide = "biopsy_wide"
    clinical = "clinical"
    screening = "screening"
    surgery = "surgery"


@dataclass
class BaseEvent:
    """
    Generic NBSS information pertaining to a medical procedure.

    :param left_opinion: Code summarising the diagnosis of the left breast
    :param right_opinion: Code summarising the diagnosis of the right breast
    :param dates: Dates extracted from left and right lesion data
    """

    left_opinion: Optional[SideOpinion] = None
    right_opinion: Optional[SideOpinion] = None
    dates: Optional[List[datetime.date]] = None


@dataclass
class BreastScreeningData:
    """
    Side-specific breast-screening data.

    :param date: Date that the examination took place
    :param equipment_make_model: Short description of the equipment used for
        the examination. Typically the manufacturer of the imaging device.
    :param opinion: Enumeration representing the opinion code
    """

    date: Optional[datetime.date] = None
    equipment_make_model: Optional[str] = None
    opinion: Optional[Opinion] = None


@dataclass
class Screening(BaseEvent):
    """
    NBSS information pertaining to a screening procedure.

    :param left: Screening information for the left breast
    :param right: Screening information for the right breast
    """

    left: Optional[BreastScreeningData] = None
    right: Optional[BreastScreeningData] = None


@dataclass
class Events:
    """
    A collection of events, for a given :class:`omidb.core.Episode`.
    """

    screening: Optional[List[Screening]] = None
    assessment: Optional[BaseEvent] = None
    clinical: Optional[BaseEvent] = None
    biopsy_wide: Optional[BaseEvent] = None
    biopsy_fine: Optional[BaseEvent] = None
    surgery: Optional[BaseEvent] = None



import pytest
import datetime
from typing import Optional, Union
from omidb import episode, events
from omidb.classificationtools import (
    episode_outcome,
    is_post_op,
    _within_time_period,
    _is_time_bounded_ci_prior,
    _is_time_bounded_cancer_prior,
    _find_next_episode,
    EpisodeOutcome,
    UndefinedEpisodeOutcome,
)


def test_find_next_episode() -> None:
    episodes = [
        episode.Episode(
            id="1",
            events=events.Events(
                screening=[events.Screening(dates=[datetime.date(2000, 2, 1)])]
            ),
        ),
        episode.Episode(
            id="2",
            events=events.Events(
                screening=[events.Screening(dates=[datetime.date(2000, 1, 1)])]
            ),
        ),
    ]

    next_epsiode = _find_next_episode(episodes[0], episodes)
    assert next_epsiode is None
    next_epsiode = _find_next_episode(episodes[1], episodes)
    assert next_epsiode == episodes[0]


@pytest.mark.parametrize("threshold, expected", [(2, True), (1, False)])
def test_within_time_period(threshold: int, expected: bool) -> None:
    e1 = episode.Episode(
        id="1",
        events=events.Events(
            screening=[events.Screening(dates=[datetime.date(2000, 1, 1)])]
        ),
    )
    e2 = episode.Episode(
        id="2",
        events=events.Events(
            screening=[events.Screening(dates=[datetime.date(2000, 3, 1)])]
        ),
    )
    assert _within_time_period(e1, e2, threshold) == expected


@pytest.mark.parametrize(
    "opinion, expected", [(events.SideOpinion.OM, True),
                          (events.SideOpinion.OS, None)]
)
def test_is_time_bounded_ci_prior(opinion: events.SideOpinion, expected: bool) -> None:
    e1 = episode.Episode(
        id="1",
        events=events.Events(
            screening=[events.Screening(dates=[datetime.date(2000, 1, 1)])]
        ),
    )
    e2 = episode.Episode(
        id="2",
        events=events.Events(
            surgery=events.BaseEvent(
                dates=[datetime.date(2000, 3, 1)], left_opinion=opinion
            )
        ),
        type=episode.Type.CI,
    )

    assert _is_time_bounded_ci_prior(e1, e2) == expected


@pytest.mark.parametrize(
    "opinion, expected", [(events.SideOpinion.OM, True),
                          (events.SideOpinion.OS, False)]
)
def test_is_time_bounded_cancer_prior(
    opinion: events.SideOpinion, expected: bool
) -> None:
    e1 = episode.Episode(
        id="1",
        events=events.Events(
            screening=[events.Screening(dates=[datetime.date(2000, 1, 1)])]
        ),
    )
    e2 = episode.Episode(
        id="2",
        events=events.Events(
            surgery=events.BaseEvent(
                dates=[datetime.date(2000, 3, 1)], left_opinion=opinion
            )
        ),
    )

    assert _is_time_bounded_cancer_prior(e1, e2) == expected


@pytest.mark.parametrize("days_offset, expected", [(-1, False), (0, False), (1, True)])
def test_is_post_op(days_offset: int, expected: bool) -> None:
    surgery_date = datetime.date(2000, 2, 1)
    screening_date = surgery_date + datetime.timedelta(days=days_offset)
    episodes = [
        episode.Episode(
            id="1",
            events=events.Events(
                screening=[events.Screening(dates=[screening_date])]),
        ),
        episode.Episode(
            id="2",
            events=events.Events(
                surgery=events.BaseEvent(
                    dates=[surgery_date], left_opinion=events.SideOpinion.OM
                )
            ),
        ),
    ]

    assert is_post_op(episodes[0], episodes) == expected


@pytest.mark.parametrize(
    "ep_type, opinion, expected",
    [
        (episode.Type.CI, events.SideOpinion.OM, True),
        (episode.Type.CI, events.SideOpinion.OS, False),
        (episode.Type.F, events.SideOpinion.OM, False),
    ],
)
def test_is_interval_cancer_surgery_opinion(
    ep_type: episode.Type, opinion: events.SideOpinion, expected: bool
) -> None:
    e = episode.Episode(
        id="1",
        events=events.Events(
            surgery=events.BaseEvent(
                dates=[datetime.date(2000, 3, 1)], left_opinion=opinion
            )
        ),
        type=ep_type,
    )

    assert e.is_interval_cancer == expected


def test_is_interval_cancer_no_events_no_dates() -> None:
    e = episode.Episode(id="1", events=None, type=episode.Type.CI)
    assert e.is_interval_cancer


def test_classify_no_events() -> None:
    e = episode.Episode(id="1", events=None, type=episode.Type.R)
    assert episode_outcome(e, [e])[0] == UndefinedEpisodeOutcome.InvalidEvents


def test_classify_no_dates() -> None:
    e = episode.Episode(
        id="1",
        events=events.Events(
            screening=events.BaseEvent(
                dates=None,
                left_opinion=events.SideOpinion.ON,
            )
        ),
        type=episode.Type.R,
        opened_date=None,
    )
    assert episode_outcome(e, [e])[0] == UndefinedEpisodeOutcome.DateError


def test_classify_next_episode_no_events() -> None:

    episodes = [
        episode.Episode(
            id="1",
            type=episode.Type.F,
            events=events.Events(
                screening=events.BaseEvent(
                    dates=[datetime.date(2000, 3, 1)],
                    left_opinion=events.SideOpinion.ON,
                )
            ),
        ),
        episode.Episode(
            id="2",
            type=episode.Type.R,
            events=None,
            opened_date=datetime.date(2000, 4, 1),
        ),
    ]

    expected = [
        UndefinedEpisodeOutcome.InvalidEvents,
        UndefinedEpisodeOutcome.InvalidEvents
    ]
    for ep, exp in zip(episodes, expected):
        assert episode_outcome(
            ep, episodes, num_months_normal_follow_up=0)[0] == exp


def test_classify_null_events() -> None:
    episodes = [
        episode.Episode(
            id="1",
            events=events.Events(
                screening=[events.Screening()],
                assessment=events.BaseEvent(
                    left_opinion=events.SideOpinion.ON),
                biopsy_fine=events.BaseEvent(
                    left_opinion=events.SideOpinion.ON),
            ),
            opened_date=datetime.date(2000, 1, 1),
        ),
        episode.Episode(  # This is the problematic episode
            id="2",
            events=events.Events(
                screening=[events.Screening()],
                assessment=events.BaseEvent(),
            ),
            opened_date=datetime.date(2000, 2, 1),
        ),
    ]

    assert (
        episode_outcome(episodes[0], episodes)[0]
        == UndefinedEpisodeOutcome.InvalidEvents
    )
    assert (
        episode_outcome(episodes[1], episodes)[0]
        == UndefinedEpisodeOutcome.InvalidEvents
    )


def test_classify_null_events_with_time_threshold() -> None:
    episodes = [
        episode.Episode(
            id="1",
            events=events.Events(
                screening=[events.Screening()],
                assessment=events.BaseEvent(
                    left_opinion=events.SideOpinion.ON),
                biopsy_fine=events.BaseEvent(
                    left_opinion=events.SideOpinion.ON),
            ),
            opened_date=datetime.date(2000, 1, 1),
        ),
        episode.Episode(  # This is fine
            id="2",
            events=events.Events(
                screening=[events.Screening()],
            ),
            opened_date=datetime.date(2000, 2, 1),
        ),
        episode.Episode(  # Problematic episode within the time window
            id="3",
            events=events.Events(
                screening=[events.Screening()],
                assessment=events.BaseEvent(),
            ),
            opened_date=datetime.date(2000, 3, 1),
        ),
    ]

    assert (
        episode_outcome(episodes[0], episodes, 10, 10)[0]
        == UndefinedEpisodeOutcome.InvalidEvents
    )


def test_classify_normals() -> None:

    episodes = [
        episode.Episode(
            id="1",
            events=events.Events(
                screening=[events.Screening(
                    dates=[datetime.date(2000, 1, 1)])],
                assessment=events.BaseEvent(
                    left_opinion=events.SideOpinion.ON,
                ),
                biopsy_fine=events.BaseEvent(
                    left_opinion=events.SideOpinion.ON,
                ),
            ),
        ),
        episode.Episode(
            id="2",
            events=events.Events(
                screening=[events.Screening(
                    dates=[datetime.date(2000, 2, 1)])],
                assessment=events.BaseEvent(
                    left_opinion=events.SideOpinion.ON,
                ),
            ),
        ),
        episode.Episode(
            id="3",
            events=events.Events(
                screening=[events.Screening(
                    dates=[datetime.date(2000, 3, 1)])],
            ),
        ),
        episode.Episode(  # Null date but treated as valid
            id="4",
            events=events.Events(
                screening=[events.Screening()],
            ),
            opened_date=datetime.date(2000, 4, 1),
        ),
        episode.Episode(
            id="5",
            events=events.Events(
                screening=[events.Screening(
                    dates=[datetime.date(2000, 4, 2)])],
            ),
        ),
    ]

    expected = [
        EpisodeOutcome.NAB,
        EpisodeOutcome.NA,
        EpisodeOutcome.N,
        EpisodeOutcome.N,
        UndefinedEpisodeOutcome.NoSubsequentEpisode,
    ]
    for ep, exp in zip(episodes, expected):
        assert episode_outcome(
            ep,
            episodes,
            num_months_normal_follow_up=0,
        )[0] == exp


def test_classify_cancers() -> None:

    episodes = [
        episode.Episode(
            id="1",
            events=events.Events(
                surgery=events.BaseEvent(
                    dates=[datetime.date(2000, 3, 1)],
                    left_opinion=events.SideOpinion.OM,
                )
            ),
        ),
        episode.Episode(
            id="2",
            type=episode.Type.CI,
            events=events.Events(
                surgery=events.BaseEvent(
                    dates=[datetime.date(2000, 3, 1)],
                    left_opinion=events.SideOpinion.OM,
                )
            ),
        ),
    ]

    expected = [
        EpisodeOutcome.M,
        EpisodeOutcome.CI,
    ]
    for ep, exp in zip(episodes, expected):
        assert episode_outcome(ep, episodes)[0] == exp


def test_ci_prior_no_events() -> None:
    date = datetime.date(2000, 1, 1)
    date2 = date + datetime.timedelta(days=1)

    episodes = [
        episode.Episode(
            id="1",
            events=events.Events(
                screening=[events.Screening(dates=[date])],
            ),
        ),
        episode.Episode(
            id="2", events=None, type=episode.Type.CI, diagnosis_date=date2
        ),
    ]

    assert episode_outcome(episodes[0], episodes)[0] == EpisodeOutcome.CIP
    assert episode_outcome(episodes[1], episodes)[0] == EpisodeOutcome.CI


@pytest.mark.parametrize(
    "ep_type, expected",
    [
        (episode.Type.R, EpisodeOutcome.MP),
        (episode.Type.CI, EpisodeOutcome.CIP),
    ],
)
def test_cancer_prior_default_thresholds(
    ep_type: episode.Type, expected: EpisodeOutcome
) -> None:

    date = datetime.date(2000, 1, 1)
    date2 = date + datetime.timedelta(days=10 * 365 / 12)
    date3 = date + datetime.timedelta(days=100 * 365 / 12)

    episodes = [
        episode.Episode(
            id="1",
            events=events.Events(
                screening=[events.Screening(dates=[date])],
            ),
        ),
        episode.Episode(
            id="2",
            events=events.Events(
                screening=[events.Screening(dates=[date])],
            ),
        ),
        episode.Episode(
            id="3",
            events=events.Events(
                surgery=events.BaseEvent(
                    dates=[date2], left_opinion=events.SideOpinion.OM
                )
            ),
            type=ep_type,
        ),
    ]

    assert episode_outcome(episodes[0], episodes)[0] == EpisodeOutcome.N
    assert episode_outcome(episodes[1], episodes)[0] == expected


@pytest.mark.parametrize(
    "month_offset, ep_type, expected",
    [
        (39, episode.Type.R, EpisodeOutcome.MP),
        (40, episode.Type.R, EpisodeOutcome.N),
        (36, episode.Type.CI, EpisodeOutcome.CIP),
        (37, episode.Type.CI, UndefinedEpisodeOutcome.InvalidPrior),
    ],
)
def test_cancer_prior_with_months(
    month_offset: int, ep_type: episode.Type, expected: Optional[EpisodeOutcome]
) -> None:

    date = datetime.date(2000, 1, 1)
    date2 = date + datetime.timedelta(days=(month_offset - 1) * 365 / 12)
    date3 = date + datetime.timedelta(days=month_offset * 365 / 12)

    episodes = [
        episode.Episode(
            id="1",
            events=events.Events(
                screening=[events.Screening(dates=[date])],
            ),
        ),
        episode.Episode(
            id="2",
            events=events.Events(
                screening=[events.Screening(dates=[date2])],
            ),
        ),
        episode.Episode(
            id="3",
            events=events.Events(
                surgery=events.BaseEvent(
                    dates=[date3], left_opinion=events.SideOpinion.OM
                )
            ),
            type=ep_type,
        ),
    ]

    got = episode_outcome(
        episodes[0], episodes, num_months_ci_prior=36, num_months_cancer_prior=39
    )[0]
    assert got == expected


@pytest.mark.parametrize(
    "cancer_interval, cancer_month_offset, normal_month_offset, expected",
    [
        (20, 21, 31, UndefinedEpisodeOutcome.InvalidPrior),
        (20, 32, 31, EpisodeOutcome.N),
        (None, 1, 31, EpisodeOutcome.MP),
        (None, 32, 31, EpisodeOutcome.N),
        (None, 32, 30, UndefinedEpisodeOutcome.InvalidPrior),
    ],
)
def test_normals_follow_up_and_cancer_prior_rules(
    cancer_interval: Optional[int],
    cancer_month_offset: int,
    normal_month_offset: int,
    expected: Optional[EpisodeOutcome],
) -> None:

    date = datetime.date(2000, 1, 1)
    # not an MP as too late
    date2 = date + datetime.timedelta(days=cancer_month_offset * 365 / 12)
    # satisfies normal follow up requirement
    date3 = date + datetime.timedelta(days=normal_month_offset * 365 / 12)

    episodes = [
        episode.Episode(
            id="1",
            events=events.Events(
                screening=[events.Screening(dates=[date])],
            ),
        ),
        episode.Episode(
            id="2",
            events=events.Events(
                surgery=events.BaseEvent(
                    dates=[date2], left_opinion=events.SideOpinion.OM
                )
            ),
        ),
        episode.Episode(
            id="3",
            events=events.Events(
                screening=[events.Screening(dates=[date3])],
            ),
        ),
    ]
    got = episode_outcome(
        episodes[0],
        episodes,
        num_months_cancer_prior=cancer_interval,
        num_months_normal_follow_up=30,
    )[0]
    assert got == expected


@pytest.mark.parametrize(
    "offset1, offset2, benign_follow_up_months, expected",
    [
        (39, 80, None, EpisodeOutcome.MP),
        (40, 80, None, EpisodeOutcome.B),
        (40, 80, 1, UndefinedEpisodeOutcome.InvalidPrior),
        (45, 80, 60, UndefinedEpisodeOutcome.InvalidPrior),
        (-1, -1, None, EpisodeOutcome.B),  # no follow-up
        (-1, 2, 1, EpisodeOutcome.B),  # normal follow up
        # normal follow up too early
        (-1, 2, 3, UndefinedEpisodeOutcome.InvalidFollowUp),
        # follow up too early, cancer too later
        (80, 2, 3, UndefinedEpisodeOutcome.InvalidPrior),
    ],
)
def test_benign(
    offset1: int,
    offset2: int,
    benign_follow_up_months: Optional[int],
    expected: Union[EpisodeOutcome, UndefinedEpisodeOutcome],
) -> None:

    date = datetime.date(2000, 1, 1)
    date2 = date + datetime.timedelta(days=offset1 * 365 / 12)
    date3 = date + datetime.timedelta(days=offset2 * 365 / 12)

    episodes = [
        episode.Episode(
            id="1",
            events=events.Events(
                biopsy_wide=events.BaseEvent(
                    dates=[date], left_opinion=events.SideOpinion.OB),
            ),
        ),
        episode.Episode(
            id="2",
            events=events.Events(
                surgery=events.BaseEvent(
                    dates=[date2], left_opinion=events.SideOpinion.OM
                )
            ),
        ),
        episode.Episode(
            id="3",
            events=events.Events(
                screening=[events.Screening(dates=[date3])],
            ),
        ),
    ]
    got = episode_outcome(
        episodes[0],
        episodes,
        num_months_cancer_prior=39,
        num_months_ci_prior=39,
        num_months_benign_follow_up=benign_follow_up_months,
    )[0]
    assert got == expected


def test_invalid_follow_up():

    date = datetime.date(2000, 1, 1)
    date2 = date + datetime.timedelta(days=2 * 365 / 12)

    episodes = [
        episode.Episode(
            id="1",
            events=events.Events(
                screening=[events.Screening(dates=[date])],
            ),
        ),
        episode.Episode(
            id="2",
            events=events.Events(
                screening=[events.Screening(dates=[date2])],
            ),
        ),
    ]

    got = episode_outcome(
        episodes[0],
        episodes,
        num_months_normal_follow_up=2,
    )[0]

    assert got == UndefinedEpisodeOutcome.InvalidFollowUp



from typing import List, Iterable, Tuple, Union, Any, Optional, Iterator
from dataclasses import dataclass
import dataclasses
import csv
import click
from .. import utilities
from ..client import Client
from ..image import Image
from ..study import Study
from ..series import Series
from ..episode import Episode
from .. import classificationtools as ct
from ..parser import DB
from loguru import logger
import datetime


@dataclass
class Config:
    db: str
    output_file: str
    link_all: bool
    clients_file: str
    log_file: str
    nbss_dir: str
    num_months_cancer_prior: Optional[int]
    num_months_ci_prior: Optional[int]
    num_months_normal_follow_up: Optional[int]
    num_months_benign_follow_up: Optional[int]


def flatten(
    clients: Iterable[Client],
    config: Config,
) -> Iterator[
    Tuple[
        Client,
        Episode,
        Optional[datetime.date],
        Optional[str],
        Optional[str],
        Optional[bool],
        Optional[Study],
        Optional[Series],
        Optional[Image],
    ]
]:
    for client in clients:
        for episode in client.episodes:
            episode_sort_date: Optional[datetime.date]
            try:
                episode_sort_date = ct._earliest_date(episode)
            except Exception:
                logger.exception(
                    f"Failed to extract sort date for {client.id} / {episode.id}"
                )
                episode_sort_date = None

            episode_outcome: Optional[str] = None
            related_episode_id: Optional[str] = None
            try:
                outcome, related_episode_id = ct.episode_outcome(
                    episode=episode,
                    all_episodes=client.episodes,
                    num_months_ci_prior=config.num_months_ci_prior,
                    num_months_cancer_prior=config.num_months_cancer_prior,
                    num_months_normal_follow_up=config.num_months_normal_follow_up,
                    num_months_benign_follow_up=config.num_months_benign_follow_up,
                )
                if outcome is not None:
                    episode_outcome = outcome.name
            except Exception:
                logger.exception(f"Failed to classify {client.id} / {episode.id}")

            post_op: Optional[bool] = None
            try:
                post_op = ct.is_post_op(episode, client.episodes)
            except Exception:
                logger.exception(
                    "Failed to identify 'post_op' status for "
                    f"{client.id} / {episode.id}"
                )
                post_op = None

            if not episode.studies:
                yield (
                    client,
                    episode,
                    episode_sort_date,
                    episode_outcome,
                    related_episode_id,
                    post_op,
                    None,
                    None,
                    None,
                )
            else:
                for study in episode.studies:
                    for series in study.series:
                        for image in series.images:
                            yield (
                                client,
                                episode,
                                episode_sort_date,
                                episode_outcome,
                                related_episode_id,
                                post_op,
                                study,
                                series,
                                image,
                            )


@dataclass
class DicomAttributes:
    Manufacturer: Optional[str] = None
    Model: Optional[str] = None
    PresentationIntentType: Optional[str] = None
    ImageLaterality: Optional[str] = None
    ViewPosition: Optional[str] = None
    ViewModCodeValue: Optional[str] = None
    ViewModCodeMeaning: Optional[str] = None
    BodyPartThicknessMM: Optional[float] = None
    PatientAgeYears: Optional[int] = None
    TransferSyntaxUID: Optional[str] = None


def extract_dicom_attributes(image: Image) -> DicomAttributes:
    """Forcefully extract dicom attributes"""

    manufacturer = utilities.try_image_attribute(image, "00080070")
    model = utilities.try_image_attribute(image, "00081090")
    presentation_intent_type = utilities.try_image_attribute(image, "00080068")
    image_laterality = utilities.try_image_attribute(image, "00200062")
    view_position = utilities.try_image_attribute(image, "00185101")
    view_mod_code_value = utilities.try_image_attribute(image, ["00540220", "00080100"])
    view_mod_code_meaning = utilities.try_image_attribute(
        image, ["00540220", "00080104"]
    )

    thickness = utilities.try_image_attribute(image, "001811A0")
    age = utilities.parse_age(image)
    transfer_syntax_uid = utilities.try_image_attribute(image, "00020010")

    return DicomAttributes(
        Manufacturer=manufacturer,
        Model=model,
        PresentationIntentType=presentation_intent_type,
        ImageLaterality=image_laterality,
        ViewPosition=view_position,
        ViewModCodeValue=view_mod_code_value,
        ViewModCodeMeaning=view_mod_code_meaning,
        BodyPartThicknessMM=thickness,
        PatientAgeYears=age,
        TransferSyntaxUID=transfer_syntax_uid,
    )


class DataWriter:
    def __init__(self) -> None:
        # Set up header
        self.rows: List[List[Any]] = [
            [
                "ClientID",
                "Site",
                "EpisodeID",
                "EpisodeSortDate",
                "EpisodeStatus",
                "EpisodeOutcome",
                "EpisodeOutcomeFutureEpisodeID",
                "EpisodeIsPostOp",
                "EpisodeType",
                "EpisodeAction",
                "EpisodeContainsMalignantOpinions",
                "EpisodeContainsBenignOpinions",
                "EpisodeOpenedDate",
                "EpisodeClosedDate",
                "ActualEpisodeOpenedYear",
                "EpisodeHasEvents",
                "StudyInstanceUID",
                "StudyDate",
                "EventType",
                "SeriesInstanceUID",
                "SOPInstanceUID",
                "NumberOfMarks",
            ]
        ]

        for field in dataclasses.fields(DicomAttributes):
            self.rows[0].append(field.name)

    def add(
        self,
        client: Client,
        episode: Episode,
        episode_sort_date: Optional[datetime.date],
        episode_outcome: Optional[str],
        episode_outcome_future_id: Optional[str],
        post_op: Optional[bool],
        study: Optional[Study],
        series: Optional[Series],
        image: Optional[Image],
    ) -> None:
        event_types: List[Union[None, str]] = [None]

        if study and study.event_type:
            event_types = [e.name for e in study.event_type]

        if image:
            tags = extract_dicom_attributes(image)
            num_marks = len(image.marks) if image.marks else 0
        else:
            tags = DicomAttributes()  # type: ignore
            num_marks = 0

        episode_status = episode.status.name if episode.status else None

        # 1 row per event
        for event_type in event_types:
            ep_action = episode.action.name if episode.action else None
            ep_type = episode.type.name if episode.type else None

            row = [
                client.id,
                client.site,
                episode.id,
                episode_sort_date,
                episode_status,
                episode_outcome,
                episode_outcome_future_id,
                post_op,
                ep_type,
                ep_action,
                episode.has_malignant_opinions,
                episode.has_benign_opinions,
                episode.opened_date,
                episode.closed_date,
                episode.actual_opened_year,
                episode.events is not None,
                study.id if study else None,
                study.date if study else None,
                event_type,
                series.id if series else None,
                image.id if image else None,
                num_marks,
            ]

            for tag in self.rows[0][len(row) :]:
                row.append(getattr(tags, tag))

            self.rows.append(row)

    def write(self, outfile: str) -> None:
        with open(outfile, "w") as f:
            writer = csv.writer(f, delimiter=",")
            writer.writerows(self.rows)


def run(config: Config) -> None:
    client_list = None
    if config.clients_file:
        with open(config.clients_file, "r") as f:
            client_list = [_.strip() for _ in f.readlines()]

    db = DB(
        config.db,
        clients=client_list,
        distinct_event_study_links=(not config.link_all),
        nbss_dir=config.nbss_dir,
    )
    writer = write_client_data(db, config)
    writer.write(config.output_file)


def write_client_data(
    clients: Iterable[Client], config: Optional[Config] = None
) -> DataWriter:
    if config is None:
        config = Config("", "", False, "", "", "", None, None, None, None)
    writer: DataWriter = DataWriter()
    for data in flatten(clients, config):
        writer.add(*data)
    return writer


@click.command("summarise")
@click.argument("db", type=click.Path(exists=True))
@click.argument("output-file", type=click.Path(exists=False))
@click.option(
    "--link-all",
    is_flag=True,
    help="Link all images to events, even if links are not unique",
)
@click.option(
    "--clients-file",
    type=click.Path(exists=True),
    help="File containing a list of clients to parse",
)
@click.option("--log-file", type=click.Path(exists=False), help="Log to this file")
@click.option(
    "--nbss-dir",
    type=click.Path(exists=False),
    help="Path to alternative directory where NBSS files can be found",
)
@click.option(
    "--num-months-cancer-prior",
    type=int,
    default=None,
    help=(
        "The maximum number of months between a malignant episode and any previous "
        "episode for a previous episode to be considered a prior."
    ),
)
@click.option(
    "--num-months-ci-prior",
    type=int,
    default=None,
    help=(
        "The maximum number of months between an interval cancer and any previous "
        "episode for a previous episode to be considered a prior."
    ),
)
@click.option(
    "--num-months-normal-follow-up",
    type=int,
    default=None,
    help="The number of months after which a second non-malignant episode must exist",
)
@click.option(
    "--num-months-benign-follow-up",
    type=int,
    default=None,
    help="The number of months after which a second non-malignant episode must exist",
)
def cli(
    db: str,
    output_file: str,
    link_all: bool,
    clients_file: str,
    log_file: str,
    nbss_dir: str,
    num_months_cancer_prior: Optional[int],
    num_months_ci_prior: Optional[int],
    num_months_normal_follow_up: Optional[int],
    num_months_benign_follow_up: Optional[int],
) -> None:
    """ "Write a csv file, OUTPUT_FILE, summarising the content of OMI-DB,
    located at DB
    """

    config = Config(
        db,
        output_file,
        link_all,
        clients_file,
        log_file,
        nbss_dir,
        num_months_cancer_prior,
        num_months_ci_prior,
        num_months_normal_follow_up,
        num_months_benign_follow_up,
    )

    logger.enable("omidb")

    if log_file:
        logger.remove()
        logger.add(log_file)

    run(config)



from typing import List
import datetime
import omidb
from omidb.classificationtools import has_prior


def create_client(years: List[int], ascending: bool = True) -> omidb.client.Client:

    events = omidb.events.Events(
        surgery=omidb.events.BaseEvent(left_opinion=omidb.events.SideOpinion.OM)
    )

    episodes = [
        omidb.episode.Episode(
            "a",
            omidb.events.Events(),
            None,
            opened_date=datetime.date(2013, 1, 1),  # type: ignore
        ),
        omidb.episode.Episode(
            "b",
            omidb.events.Events(),
            None,
            opened_date=datetime.date(2014, 1, 1),  # type: ignore
        ),
        omidb.episode.Episode(
            "c",
            omidb.events.Events(),
            None,
            opened_date=datetime.date(2015, 1, 1),  # type: ignore
        ),
    ]

    if not ascending:
        episodes.reverse()

    for year in years:
        for episode in episodes:
            if episode.opened_date.year == year:  # type: ignore
                episode.events = events

    return omidb.client.Client("abc1", episodes, "site")


def test_has_prior_event() -> None:

    for ascending in (True, False):

        client = create_client([2013], ascending)
        assert not has_prior(client)

        client = create_client([2014], ascending)
        assert has_prior(client)

        client = create_client([2015], ascending)
        assert has_prior(client)

        client = create_client([2013, 2014], ascending)
        assert not has_prior(client)

        client = create_client([2013, 2015], ascending)
        assert has_prior(client)



import omidb


def test_episode_has_benign_opinions_false() -> None:

    for opinion in (
        omidb.events.SideOpinion.OM,
        omidb.events.SideOpinion.ON,
        omidb.events.SideOpinion.OS,
        omidb.events.SideOpinion.OU,
    ):

        events = omidb.events.Events(
            surgery=omidb.events.BaseEvent(left_opinion=opinion),
        )

        e = omidb.episode.Episode("a", events, None)  # type: ignore
        assert not e.has_benign_opinions


def test_episode_has_malignant_opinions_false() -> None:

    for opinion in (
        omidb.events.SideOpinion.OB,
        omidb.events.SideOpinion.ON,
        omidb.events.SideOpinion.OS,
        omidb.events.SideOpinion.OU,
    ):

        events = omidb.events.Events(
            surgery=omidb.events.BaseEvent(left_opinion=opinion),
        )

        e = omidb.episode.Episode("a", events, None)  # type: ignore
        assert not e.has_malignant_opinions


def test_episode_has_benign_opinions_true() -> None:

    event_ob = omidb.events.BaseEvent(left_opinion=omidb.events.SideOpinion.OB)
    event_om = omidb.events.BaseEvent(left_opinion=omidb.events.SideOpinion.OM)
    for event1, event2 in ((event_ob, event_om), (event_om, event_ob)):
        events = omidb.events.Events(surgery=event1, biopsy_wide=event2)
        e = omidb.episode.Episode("a", events, None)  # type: ignore
        assert e.has_benign_opinions


def test_episode_has_malignant_opinions_true() -> None:

    event_ob = omidb.events.BaseEvent(left_opinion=omidb.events.SideOpinion.OB)
    event_om = omidb.events.BaseEvent(left_opinion=omidb.events.SideOpinion.OM)

    for event1, event2 in ((event_ob, event_om), (event_om, event_ob)):
        events = omidb.events.Events(surgery=event1, biopsy_wide=event2)
        e = omidb.episode.Episode("a", events, None)  # type: ignore
        assert e.has_malignant_opinions


def test_episode_has_benign_opinions_side() -> None:

    left = omidb.events.BaseEvent(left_opinion=omidb.events.SideOpinion.OB)
    right = omidb.events.BaseEvent(right_opinion=omidb.events.SideOpinion.OB)

    for side in (left, right):

        events_surgery = omidb.events.Events(surgery=side)
        events_biopsy = omidb.events.Events(biopsy_wide=side)

        for events in (events_surgery, events_biopsy):
            e = omidb.episode.Episode("a", events, None)  # type: ignore
            assert e.has_benign_opinions


def test_episode_has_malignant_opinions_side() -> None:

    left = omidb.events.BaseEvent(left_opinion=omidb.events.SideOpinion.OM)
    right = omidb.events.BaseEvent(right_opinion=omidb.events.SideOpinion.OM)

    for side in (left, right):

        events_surgery = omidb.events.Events(surgery=side)
        events_biopsy = omidb.events.Events(biopsy_wide=side)

        for events in (events_surgery, events_biopsy):
            e = omidb.episode.Episode("a", events, None)  # type: ignore
            assert e.has_malignant_opinions



from __future__ import annotations
from typing import Dict, List, Callable, Iterator, Union
from .client import Client
from .episode import Episode
from .study import Study
from .series import Series
from .image import Image


class FilterImages:
    """
    Filter images of a hierarchical Image model, by applying a user-specified
    filtering function to each image.

    :param image_filter: Use `image_filter` to filter images provided by the caller.
    """

    def __init__(self, image_filter: Callable[[Image], bool]):
        self.image_filter = image_filter

    @classmethod
    def dicom_filter(cls, tag_criteria: Dict[str, List[str]]) -> FilterImages:
        """
        Returns a new :class:`FilterImages` instance whose filter function uses the
        tag/attribute name of a Data Element to filter images by the value of a
        specific Data Element, i.e. filter by dicom property.

        :param tag_criteria: Dictionary whose keys are the data element tags
            and values are a list of data element values
        """

        def the_filter(image: Image) -> bool:
            ds = image.dcm

            if ds is None:
                raise ValueError(f"Failed to load DICOM for image {image.id}")

            for tag, value in tag_criteria.items():
                v = ds.data_element(tag)
                if v is not None and v.value not in value:
                    return False

            return True

        return cls(the_filter)

    def _from_client_iter(self, client: Client) -> Iterator[Episode]:
        """
        Iterate over episodes within `client`
        """
        for episode in client.episodes:
            for _ in self._from_episode_iter(episode):
                yield (_)

    def _from_episode_iter(self, episode: Episode) -> Iterator[Study]:
        """
        Iterate over studies within `episode`
        """
        if episode.studies:
            for study in episode.studies:
                for _ in self._from_study_iter(study):
                    yield (_)

    def _from_study_iter(self, study: Study) -> Iterator[Series]:
        """
        Iterate over series within `study`
        """
        for series in study.series:
            yield (series)

    def __call__(self, obj: Union[Client, Episode, Study, Series]) -> None:
        """
        Inplace filtering of the :class:`omidb.image.Image` s found in ``obj``.

        :param obj: Filter the images nested within one of the core OMI-DB objects
        """

        if isinstance(obj, Series):
            obj.images = [_ for _ in filter(self.image_filter, obj.images)]

        elif isinstance(obj, Client):
            for _ in self._from_client_iter(obj):
                self(_)

        elif isinstance(obj, Episode):
            for _ in self._from_episode_iter(obj):
                self(_)

        elif isinstance(obj, Study):
            for _ in self._from_study_iter(obj):
                self(_)
